# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main

from pybill.lib import ACCURACY
from pybill.lib.entities.accounting_docs import PreviousAccountingDoc
from pybill.lib.entities.accounting_docs import Bill

from generic_test_accounting_docs import AbstractAccDocTest


class BillTest(TestCase, AbstractAccDocTest):
    """
    Tests on the bill objects.
    """
    def setUp(self):
        """
        Called before each test from this class.
        """
        self.acc_doc = Bill("ID1")

    def define_downpayments_in_example(self):
        dwp = PreviousAccountingDoc()
        dwp.total = 23.0
        dwp.vat_amount = 2.0
        self.acc_doc.charged_downpayments.append(dwp)
        dwp = PreviousAccountingDoc()
        dwp.total = 13.0
        dwp.vat_amount = 1.0
        self.acc_doc.charged_downpayments.append(dwp)

    def define_debits_in_example(self):
        dbt = PreviousAccountingDoc()
        dbt.total = 27.0
        dbt.vat_amount = 3.0
        self.acc_doc.issued_debits.append(dbt)
        dbt = PreviousAccountingDoc()
        dbt.total = 17.0
        dbt.vat_amount = 1.5
        self.acc_doc.issued_debits.append(dbt)
        
    def test_get_vat_amount_for_rate_with_downpayment(self):
        """
        Tests the method giving the VAT amount for a given VAT rate when the
        list of already charged downpayments is not empty.
        """
        self.build_vat_example()
        self.define_downpayments_in_example()
        self.assertEqual(self.acc_doc.get_vat_amount_for_rate(1.0),
                          round(2.56, ACCURACY))
        self.assertEqual(self.acc_doc.get_vat_amount_for_rate(11.0),
                          round(3.85, ACCURACY))
        self.assertEqual(self.acc_doc.get_vat_amount_for_rate(21.0),
                          round(7.35, ACCURACY))
        self.assertEqual(self.acc_doc.get_vat_amount_for_rate(19.6),
                          round(0.0, ACCURACY))

    def test_get_price_total_with_downpayment(self):
        """
        Tests the computation of the tax free total when the
        list of already charged downpayments is not empty.
        """
        self.build_standard_example()
        self.define_downpayments_in_example()
        self.assertEqual(self.acc_doc.get_price_total(),
                          round(711.0, ACCURACY))

    def test_get_including_taxes_total_with_downpayment(self):
        """
        Tests the computation of the including taxes total when the
        list of already charged downpayments is not empty.
        """
        self.build_standard_example()
        self.define_downpayments_in_example()
        self.assertEqual(self.acc_doc.get_including_taxes_total(),
                          round(807.04, ACCURACY))

    def test_get_holdback_amount_with_downpayment(self):
        """
        Tests the computation of the amount of holdbacks when the list
        of already charged downpayments is not empty.
        """
        self.build_standard_example()
        self.define_downpayments_in_example()
        self.assertEqual(self.acc_doc.get_holdback_amount(),
                          round(0.0, ACCURACY))

    def test_get_holdback_vat_amount_with_downpayment(self):
        """
        Tests the computation of the amount of holdbacks on VAT when the list
        of already charged downpayments is not empty.
        """
        self.build_standard_example()
        self.define_downpayments_in_example()
        self.assertEqual(self.acc_doc.get_holdback_vat_amount(),
                          round(0.0, ACCURACY))

    def test_get_total_without_holdback_to_be_paid_with_downpayment(self):
        """
        Tests the computation of the total without holdback to be paid when
        the list of already charged downpayments is not empty.
        """
        self.build_standard_example()
        self.define_downpayments_in_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback_free(),
                          round(771.04, ACCURACY))

    def test_get_total_holdback_to_be_paid_with_downpayment(self):
        """
        Tests the computation of the total of holdbacks to be paid when the
        list of already charged downpayments is not empty.
        """
        self.build_standard_example()
        self.define_downpayments_in_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback(),
                          round(0.0, ACCURACY))

    def test_get_total_vat_to_be_paid_with_downpayment(self):
        """
        Tests the computation of the total of VAT to be paid when the
        list of already charged downpayments is not empty.
        """
        self.build_standard_example()
        self.define_downpayments_in_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_vat(),
                          round(93.04, ACCURACY))

    def test_get_total_to_be_paid_with_downpayment(self):
        """
        Tests the computation of the total to be paid when the
        list of already charged downpayments is not empty.
        """
        self.build_standard_example()
        self.define_downpayments_in_example()
        self.assertEqual(self.acc_doc.get_to_be_paid(),
                          round(771.04, ACCURACY))

    def test_get_vat_amount_for_rate_with_debit(self):
        """
        Tests the method giving the VAT amount for a given VAT rate when the
        list of already issued debits is not empty.
        """
        self.build_vat_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_vat_amount_for_rate(1.0),
                          round(2.56, ACCURACY))
        self.assertEqual(self.acc_doc.get_vat_amount_for_rate(11.0),
                          round(3.85, ACCURACY))
        self.assertEqual(self.acc_doc.get_vat_amount_for_rate(21.0),
                          round(7.35, ACCURACY))
        self.assertEqual(self.acc_doc.get_vat_amount_for_rate(19.6),
                          round(0.0, ACCURACY))

    def test_get_price_total_with_debit(self):
        """
        Tests the computation of the tax free total when the
        list of already issued debits is not empty.
        """
        self.build_standard_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_price_total(),
                          round(711.0, ACCURACY))

    def test_get_including_taxes_total_with_debit(self):
        """
        Tests the computation of the including taxes total when the
        list of already issued debits is not empty.
        """
        self.build_standard_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_including_taxes_total(),
                          round(807.04, ACCURACY))

    def test_get_holdback_amount_with_debit(self):
        """
        Tests the computation of the amount of holdbacks when the
        list of already issued debits is not empty.
        """
        self.build_standard_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_holdback_amount(),
                          round(0.0, ACCURACY))

    def test_get_holdback_vat_amount_with_debit(self):
        """
        Tests the computation of the amount of holdbacks on vat when the
        list of already issued debits is not empty.
        """
        self.build_standard_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_holdback_vat_amount(),
                          round(0.0, ACCURACY))

    def test_get_total_without_holdback_to_be_paid_with_debit(self):
        """
        Tests the computation of the total without holdback to be paid when
        the list of already issued debits is not empty.
        """
        self.build_standard_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback_free(),
                          round(763.04, ACCURACY))

    def test_get_total_holdback_to_be_paid_with_debit(self):
        """
        Tests the computation of the total of holdbacks to be paid when the
        list of already issued debits is not empty.
        """
        self.build_standard_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback(),
                          round(0.0, ACCURACY))

    def test_get_total_vat_to_be_paid_with_debit(self):
        """
        Tests the computation of the total of VAT to be paid when the
        list of already issued debits is not empty.
        """
        self.build_standard_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_vat(),
                          round(91.54, ACCURACY))

    def test_get_total_to_be_paid_with_debit(self):
        """
        Tests the computation of the total to be paid when the
        list of already issued debits is not empty.
        """
        self.build_standard_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_to_be_paid(),
                          round(763.04, ACCURACY))

    def test_get_holdback_amount_with_downpayment_debit_holdback(self):
        """
        Tests the computation of the amount of holdbacks when there
        are charged downpayment, issued debits and holdbacks.
        """
        self.build_standard_example(with_holdbacks=True)
        self.define_downpayments_in_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_holdback_amount(),
                          round(33.80, ACCURACY))

    def test_get_holdback_vat_amount_with_downpayment_debit_holdback(self):
        """
        Tests the computation of the amount of holdbacks on vat when there
        are charged downpayment, issued debits and holdbacks.
        """
        self.build_standard_example(with_holdbacks=True)
        self.define_downpayments_in_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_holdback_vat_amount(),
                          round(1.372, ACCURACY))

    def test_get_total_without_hldbk_tbp_with_downpayment_debit_holdback(self):
        """
        Tests the computation of the total without holdback to be paid when
        there are charged downpayment, issued debits and holdbacks.
        """
        self.build_standard_example(with_holdbacks=True)
        self.define_downpayments_in_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback_free(),
                          round(691.868, ACCURACY))

    def test_get_total_holdback_tbp_with_downpayment_debit_holdback(self):
        """
        Tests the computation of the total of holdbacks to be paid when there
        are charged downpayment, issued debits and holdbacks.
        """
        self.build_standard_example(with_holdbacks=True)
        self.define_downpayments_in_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback(),
                          round(35.172, ACCURACY))

    def test_get_total_vat_to_be_paid_with_downpayment_debit_holdback(self):
        """
        Tests the computation of the total of VAT to be paid when there
        are charged downpayment, issued debits and holdbacks.
        """
        self.build_standard_example(with_holdbacks=True)
        self.define_downpayments_in_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_vat(),
                          round(88.54, ACCURACY))

    def test_get_total_to_be_paid_with_downpayment_debit_holdback(self):
        """
        Tests the computation of the total to be paid when there
        are charged downpayment, issued debits and holdbacks.
        """
        self.build_standard_example(with_holdbacks=True)
        self.define_downpayments_in_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_to_be_paid(),
                          round(727.04, ACCURACY))

    def test_get_holdback_amount_with_downpayment_debit_huge_holdback(self):
        """
        Tests the computation of the amount of holdbacks when there
        are charged downpayment, issued debits and huge holdbacks.
        """
        self.build_standard_example(with_holdbacks=True)
        for it in self.acc_doc.acc_items:
            it.holdback_rate = 99.0
        self.define_downpayments_in_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_holdback_amount(),
                          round(703.89, ACCURACY))

    def test_get_holdback_vat_amount_with_downpayment_debit_huge_holdback(self):
        """
        Tests the computation of the amount of holdbacks on vat when there
        are charged downpayment, issued debits and huge holdbacks.
        """
        self.build_standard_example(with_holdbacks=True)
        for it in self.acc_doc.acc_items:
            it.holdback_rate = 99.0
        self.define_downpayments_in_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_holdback_vat_amount(),
                          round(33.957, ACCURACY))

    def test_get_total_without_hldbk_tbp_with_dwnpymt_dbt_huge_holdback(self):
        """
        Tests the computation of the total without holdback to be paid when
        there are charged downpayment, issued debits and huge holdbacks.
        """
        self.build_standard_example(with_holdbacks=True)
        for it in self.acc_doc.acc_items:
            it.holdback_rate = 99.0
        self.define_downpayments_in_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback_free(),
                          round(0.0, ACCURACY))

    def test_get_total_holdback_tbp_with_downpayment_debit_huge_holdback(self):
        """
        Tests the computation of the total of holdbacks to be paid when there
        are charged downpayment, issued debits and huge holdbacks.
        """
        self.build_standard_example(with_holdbacks=True)
        for it in self.acc_doc.acc_items:
            it.holdback_rate = 99.0
        self.define_downpayments_in_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback(),
                          round(727.04, ACCURACY))

    def test_get_total_vat_tbp_with_downpayment_debit_huge_holdback(self):
        """
        Tests the computation of the total of VAT to be paid when there
        are charged downpayment, issued debits and huge holdbacks.
        """
        self.build_standard_example(with_holdbacks=True)
        for it in self.acc_doc.acc_items:
            it.holdback_rate = 99.0
        self.define_downpayments_in_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_vat(),
                          round(88.54, ACCURACY))

    def test_get_total_to_be_paid_with_downpayment_debit_huge_holdback(self):
        """
        Tests the computation of the total to be paid when there
        are charged downpayment, issued debits and huge holdbacks.
        """
        self.build_standard_example(with_holdbacks=True)
        for it in self.acc_doc.acc_items:
            it.holdback_rate = 99.0
        self.define_downpayments_in_example()
        self.define_debits_in_example()
        self.assertEqual(self.acc_doc.get_to_be_paid(),
                          round(727.04, ACCURACY))


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-
"""
``pybill.lib.pdfwriters.flowgenerators.downpayment_generator`` module defines
the flow generator for the downpayment entities.

The flow generator generates the title, various paragraphes, the main table
detailling the items, and the total table. This flow content will be displayed
on various pages by the
:class:`~pybill.lib.pdfwriters.templates.doc_templates.AccDocDocumentTemplate`
document template with the platypus machinery.

.. autoclass:: DownpaymentFlowGenerator
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"

from pybill.lib.pdfwriters.flowgenerators.abstract import AbstractFlowGenerator
from pybill.lib.pdfwriters.utils import escape_text, format_number
from pybill.lib import ACCURACY


class DownpaymentFlowGenerator(AbstractFlowGenerator):
    """
    Class generating the PDF flow content for a
    :class:`~pybill.lib.entities.accounting_docs.downpayment.Downpayment`.

    .. automethod:: _get_acc_doc_title

    .. automethod:: _get_total_table_cell_contents
    """
    def _get_acc_doc_title(self, acc_doc):
        """
        Gives the title of the accounting document (here, a downpayment).

        :parameter acc_doc: The current accounting document that must be
                            exported in PDF.
        :type acc_doc:
            :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`

        :returns: The accounting document title.
        :rtype: :class:`unicode`
        """
        return (u"<b>%s %s %s</b> %s %s - %s" 
            % (escape_text(acc_doc.cfg_data.get_local_phrase(u'bill')),
               escape_text(acc_doc.cfg_data.get_local_phrase(u'number')),
               escape_text(acc_doc.id, True),
               escape_text(acc_doc.cfg_data.get_local_phrase(u'dated')),
               escape_text(acc_doc.date, True),
               escape_text(acc_doc.cfg_data.get_local_phrase(u'downpayment'))))

    def _get_total_table_cell_contents(self, acc_doc):
        """
        Gives the contents of the cells of the total table.

        These contents will be used by the ``_get_total_table`` method to build
        the total table.

        :parameter acc_doc: The current accounting document that must be
                            exported in PDF (here a downpayment).
        :type acc_doc:
            :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`

        :returns:
            A list describing the cells of the total table. The first element is
            a list of unicode strings corresponding to the various lines of the
            name of the row; the second element is a float corresponding to the
            number written in the row; the third element is a specification of
            the style of the row (``u"bold"``, ``u"italic"``,
            ``u"bold italic"``, ``u""``). 
        :rtype: list of triples (list of :class:`unicode`, :class:`float`,
                :class:`unicode`)
        """
        contents = []
        contents.extend(
            self._build_cell_contents_for_totals(acc_doc))
        # Next row (downpayment, i.e. amount to be really paid).
        name = [u"%s %s%%" % (acc_doc.cfg_data.get_local_phrase('downpayment'),
                              format_number(acc_doc.percent, ACCURACY,
                                          acc_doc.cfg_data.number_separators))]
        number = acc_doc.get_downpayment()
        contents.append((name, number, u"regular"))
        # Next row (amount to be paid).
        name = [acc_doc.cfg_data.get_local_phrase(u'to-be-paid')]
        number = acc_doc.get_to_be_paid_holdback_free()
        contents.append((name, number, u"bold"))
        if acc_doc.get_to_be_paid_holdback() != 0:
            name = [acc_doc.cfg_data.get_local_phrase(u'to-be-paid'), 
                    acc_doc.cfg_data.get_local_phrase(u'holdback'), ]
            number = acc_doc.get_to_be_paid_holdback()
            contents.append((name, number, u"bold"))
        return contents



# -*- coding: utf-8 -*-
"""
``pybill.lib.config.xmlreaders`` module contains the function used for reading
config XML files.

This function uses the :mod:`lxml` library to read the XML data. The
information read from one XML file is stored in a
:class:`~pybill.lib.config.entities.ConfigData` object.

.. autofunction:: load_config
"""
__docformat__ = "restructuredtext en"

from traceback import format_exc
from lxml import etree

from pybill.lib import PBC_0_X, PBC_FORMATS, PBC_XMLNS

from pybill.lib.xmlreaders.utils import (fill_address_from_address_block,
                                         read_text_from_xml_elt)
from pybill.lib.errors import PyBillReadingException
from pybill.lib.config.entities import ConfigData


def load_config(xml_file):
    """
    Loads the configuration data contained in the file ``xml_file``, stores
    it in a ``ConfigData`` object, and returns the new object.

    The ``lxml`` library is used to read the data from the XML.
        
    :parameter xml_file: name of the XML file containing the configuration
                         data.
    :type xml_file: :class:`unicode`

    :returns: A new configuration object filled with the data read in the
              XML file.
    :rtype: :class:`~pybill.lib.config.entities.ConfigData`    
    """
    config_data = ConfigData()
    # Tries to read the XML file
    try:
        xml_tree = etree.parse(xml_file)
    except Exception, exc:
        raise PyBillReadingException("Error while reading configuration "
                               "file %s\n%s" % (xml_file, format_exc(limit=1)))
    xml_root = xml_tree.getroot()
    if xml_root.tag not in  [u"config", u"{%s}config" % PBC_XMLNS]:
        raise PyBillReadingException("%s file is not a PyBill "
                                     "configuration file." % xml_file)
    # Finds config format
    cfg_format = xml_root.get(u"{%s}format-version" % PBC_XMLNS)
    if cfg_format is None and u"format-version" not in xml_root.keys():
        cfg_format = PBC_0_X
    elif cfg_format is None:
        cfg_format = xml_root.get(u"format-version")
    if cfg_format not in PBC_FORMATS:
        raise PyBillReadingException("Configuration from '%s' file is not in a"
                          " PyBill known format: %r." % (xml_file, cfg_format))
    # Sets configuration name
    cfg_name = xml_root.get(u"{%s}name" % PBC_XMLNS)
    if cfg_name is None:
        cfg_name = xml_root.get(u"name", u"")
    config_data.name = cfg_name
    # Chooses the namespace depending on the config format
    if cfg_format == PBC_0_X:
        xml_ns = u""
    else:
        xml_ns = PBC_XMLNS
    # Reads the company data
    xml_cpy = xml_root.find(u"{%s}company" % xml_ns)
    if xml_cpy is not None:
        config_data.company_logo = read_text_from_xml_elt(xml_cpy,
                                                     u"{%s}logo-file" % xml_ns)
        addr = config_data.company_address
        addr.name = read_text_from_xml_elt(xml_cpy, u"{%s}orgname" % xml_ns)
        xml_addr = xml_cpy.find(u"{%s}address" % xml_ns)
        if xml_addr is not None:
            fill_address_from_address_block(addr, xml_addr, xml_ns=xml_ns)
    # Reads the bank data
    for xml_ln in xml_root.findall(u"{%s}bank-data/{%s}line"
                                                           % (xml_ns, xml_ns)):
        line = read_text_from_xml_elt(xml_ln, u".")
        if line:
            config_data.bank_data_lines.append(unicode(line))
    # Reads the footer 
    for xml_ln in xml_root.findall(u"{%s}footer/{%s}line" % (xml_ns, xml_ns)):
        line = read_text_from_xml_elt(xml_ln, u".")
        if line:
            config_data.footer_lines.append(unicode(line))
    # Reads the agreement intro
    for xml_ln in xml_root.findall(u"{%s}agreement-intro/{%s}line"
                                                           % (xml_ns, xml_ns)):
        line = read_text_from_xml_elt(xml_ln, u".")
        if line:
            config_data.agreement_intro_lines.append(unicode(line))
    # Reads the number separators
    for sep_kw in [u"sign", u"thousands", u"digits"]:
        sep_val = read_text_from_xml_elt(xml_root,
                    u"{%s}number-separators/{%s}%s" % (xml_ns, xml_ns, sep_kw))
        if sep_val:
            config_data.number_separators[sep_kw] = sep_val
    # Reads the localisation data
    for xml_kwd in xml_root.find(u"{%s}localisation" % xml_ns):
        if xml_kwd.text and isinstance(xml_kwd.tag, str):
            kwd = unicode(xml_kwd.tag).split("}")[-1]
            config_data.local_phrases[kwd] = unicode(xml_kwd.text)
    # Finally, returns the configuration object
    return config_data



# -*- coding: utf-8 -*-
"""
``pybill.lib.config.xmlwriters`` module contains the function used for writing
``ConfigData`` objects into XML files.

This dump operation uses the :mod:`lxml` library.

.. autofunction:: save_config
"""
__docformat__ = "restructuredtext en"

from lxml import etree

from pybill.lib import PBC_1_0, PBC_XMLNS

from pybill.lib.errors import PyBillWritingException
from pybill.lib.xmlwriters.utils import (ENCODING, open_xml_outfile,
                                         write_address_xml_elements)


def save_config(cfg_data, xml_file, encoding=ENCODING):
    """
    Saves the configuration data contained in the ``cfg_data`` object into the
    ``xml_file`` file.

    The ``lxml`` library is used to write the XML data.
        
    :parameter cfg_data: Configuration object to be saved in XML.
    :type cfg_data: :class:`~pybill.lib.config.entities.ConfigData`    

    :parameter xml_file: name of the XML file that will be written.
                         data. Actually, this name is built from the reference
                         of the configuration data.
    :type xml_file: :class:`unicode`

    :parameter encoding: Encoding used to write the XML file.
    :type encoding: :class:`str`
    """
    # Builds root element
    xml_root = etree.Element(u"{%s}config" % PBC_XMLNS,
                             nsmap={"pbc": PBC_XMLNS})
    xml_root.set(u"{%s}format-version" % PBC_XMLNS, PBC_1_0)
    xml_root.set(u"{%s}name" % PBC_XMLNS, cfg_data.name)
    # Writes the company data
    xml_cpy = etree.SubElement(xml_root, u"{%s}company" % PBC_XMLNS)
    if cfg_data.company_logo.strip() != u"":
        xml_clg = etree.SubElement(xml_cpy, u"{%s}logo-file" % PBC_XMLNS)
        xml_clg.text = unicode(cfg_data.company_logo)
    xml_cnm = etree.SubElement(xml_cpy, u"{%s}orgname" % PBC_XMLNS)
    xml_cnm.text = cfg_data.company_address.name
    xml_addr = etree.Element(u"{%s}address" % PBC_XMLNS)
    write_address_xml_elements(cfg_data.company_address, xml_addr,
                               xml_ns= PBC_XMLNS)
    if xml_addr.getchildren() != []:
        xml_cpy.append(xml_addr)
    # Writes the bank data
    xml_bkd = etree.SubElement(xml_root, u"{%s}bank-data" % PBC_XMLNS)
    for line in [bkl for bkl in cfg_data.bank_data_lines
                 if bkl.strip() != u""]:
        xml_lne = etree.SubElement(xml_bkd, u"{%s}line" % PBC_XMLNS)
        xml_lne.text = line
    # Writes the footer 
    xml_ftr = etree.SubElement(xml_root, u"{%s}footer" % PBC_XMLNS)
    for line in [ftl for ftl in cfg_data.footer_lines if ftl.strip() != u""]:
        xml_lne = etree.SubElement(xml_ftr, u"{%s}line" % PBC_XMLNS)
        xml_lne.text = line
    # Writes the agreement intro
    xml_agr = etree.SubElement(xml_root, u"{%s}agreement-intro" % PBC_XMLNS)
    for line in [agl for agl in cfg_data.agreement_intro_lines
                 if agl.strip() != u""]:
        xml_lne = etree.SubElement(xml_agr, u"{%s}line" % PBC_XMLNS)
        xml_lne.text = line
    # Writes the number separators
    xml_nsp = etree.SubElement(xml_root, u"{%s}number-separators" % PBC_XMLNS)
    for sep_kw in [u"sign", u"thousands", u"digits"]:
        if cfg_data.number_separators.has_key(sep_kw):
            xml_sep = etree.SubElement(xml_nsp, u"{%s}%s" % (PBC_XMLNS, sep_kw))
            xml_sep.text = cfg_data.number_separators[sep_kw]
    # Writes the localisation data
    xml_loc = etree.SubElement(xml_root, u"{%s}localisation" % PBC_XMLNS)
    for loc_kw in [u"colon", u"phone-kw", u"fax-kw", u"web-kw", u"email-kw",
                   u"doc-ref-kw", u"receiver-kw", u"sender-kw", u"on-date",
                   u"bill", u"claim-form", u"downpayment", u"debit",
                   u"pro-forma", u"number", u"dated", u"valid-until",
                   u"intro-detail", u"quantity", u"description", u"vat-rate",
                   u"unit-price", u"tf-unit-price", u"price", u"tf-price",
                   u"it-price", u"holdback-on", u"ita-est", u"total",
                   u"tf-total", u"vat-amount", u"it-total", u"including",
                   u"holdback", u"on-tf", u"on-vat", u"debit-total",
                   u"charged-downpayment", u"charged-on", u"issued-debit",
                   u"issued-on", u"to-be-paid", u"payment-terms",
                   u"to-bring-forward", u"carry-forward", u"your-ref-kw",
                   u"purch-ref-kw", u"supplier-ref-kw"]:
        xml_lcp = etree.SubElement(xml_loc, u"{%s}%s" % (PBC_XMLNS, loc_kw))
        xml_lcp.text = cfg_data.local_phrases.get(loc_kw, u"")
    # Finally, writes the XML elements.
    out = open_xml_outfile(xml_file, encoding=encoding)
    out.write(etree.tostring(xml_root, encoding=encoding, 
                             pretty_print=True, xml_declaration=False))



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main
from lxml import etree
from pybill.lib.entities.accounting_docs import ProForma

from pybill.lib.xmlwriters.accdoc_format_1_0 import write_accdoc_1_0_xml

from generic_test_formatwriter import FormatWriterGenericTest


EXPECTED = u"""<pbd:accounting-document xmlns:pbd="http://www.logilab.org/2010/PyBillDocument" pbd:format-version="PBD-1.0" pbd:type="pro-forma">
  <pbd:validity-date>Z0</pbd:validity-date>
  <pbd:metadata>
    <pbd:doc-ref>A2</pbd:doc-ref>
    <pbd:place>A3</pbd:place>
    <pbd:date pbd:num="2009-10-19">A4</pbd:date>
    <pbd:info pbd:name="A5">A6</pbd:info>
    <pbd:info pbd:name="A7">A8</pbd:info>
    <pbd:info pbd:name="A9">A10</pbd:info>
  </pbd:metadata>
  <pbd:address pbd:role="from">
    <pbd:honorific>B1</pbd:honorific>
    <pbd:firstname>B2</pbd:firstname>
    <pbd:othername>B3</pbd:othername>
    <pbd:surname>B4</pbd:surname>
    <pbd:lineage>B5</pbd:lineage>
    <pbd:street>B6</pbd:street>
    <pbd:street>B7</pbd:street>
    <pbd:pob>B8</pbd:pob>
    <pbd:postcode>B9</pbd:postcode>
    <pbd:city>B10</pbd:city>
    <pbd:state>B11</pbd:state>
    <pbd:country>B12</pbd:country>
    <pbd:phone>B13</pbd:phone>
    <pbd:fax>B14</pbd:fax>
    <pbd:web>B15</pbd:web>
    <pbd:email>B16</pbd:email>
    <pbd:affiliation>
      <pbd:orgname>B17</pbd:orgname>
      <pbd:orgdiv>B18</pbd:orgdiv>
      <pbd:orgdiv>B19</pbd:orgdiv>
      <pbd:jobtitle>B20</pbd:jobtitle>
      <pbd:jobtitle>B21</pbd:jobtitle>
      <pbd:address>
        <pbd:street>B22</pbd:street>
        <pbd:street>B23</pbd:street>
        <pbd:pob>B24</pbd:pob>
        <pbd:postcode>B25</pbd:postcode>
        <pbd:city>B26</pbd:city>
        <pbd:state>B27</pbd:state>
        <pbd:country>B28</pbd:country>
        <pbd:phone>B29</pbd:phone>
        <pbd:fax>B30</pbd:fax>
        <pbd:web>B31</pbd:web>
        <pbd:email>B32</pbd:email>
      </pbd:address>
    </pbd:affiliation>
  </pbd:address>
  <pbd:address pbd:role="to">
    <pbd:honorific>C1</pbd:honorific>
    <pbd:firstname>C2</pbd:firstname>
    <pbd:othername>C3</pbd:othername>
    <pbd:surname>C4</pbd:surname>
    <pbd:lineage>C5</pbd:lineage>
    <pbd:street>C6</pbd:street>
    <pbd:street>C7</pbd:street>
    <pbd:pob>C8</pbd:pob>
    <pbd:postcode>C9</pbd:postcode>
    <pbd:city>C10</pbd:city>
    <pbd:state>C11</pbd:state>
    <pbd:country>C12</pbd:country>
    <pbd:phone>C13</pbd:phone>
    <pbd:fax>C14</pbd:fax>
    <pbd:web>C15</pbd:web>
    <pbd:email>C16</pbd:email>
    <pbd:affiliation>
      <pbd:orgname>C17</pbd:orgname>
      <pbd:orgdiv>C18</pbd:orgdiv>
      <pbd:orgdiv>C19</pbd:orgdiv>
      <pbd:jobtitle>C20</pbd:jobtitle>
      <pbd:jobtitle>C21</pbd:jobtitle>
      <pbd:address>
        <pbd:street>C22</pbd:street>
        <pbd:street>C23</pbd:street>
        <pbd:pob>C24</pbd:pob>
        <pbd:postcode>C25</pbd:postcode>
        <pbd:city>C26</pbd:city>
        <pbd:state>C27</pbd:state>
        <pbd:country>C28</pbd:country>
        <pbd:phone>C29</pbd:phone>
        <pbd:fax>C30</pbd:fax>
        <pbd:web>C31</pbd:web>
        <pbd:email>C32</pbd:email>
      </pbd:address>
    </pbd:affiliation>
  </pbd:address>
  <pbd:remark>D1</pbd:remark>
  <pbd:remark>D2</pbd:remark>
  <pbd:items-list>
    <pbd:item>
      <pbd:qty pbd:digits="5">501.0</pbd:qty>
      <pbd:description>
        <pbd:title>E2</pbd:title>
        <pbd:detail>E3</pbd:detail>
        <pbd:detail>E4</pbd:detail>
      </pbd:description>
      <pbd:unit-price pbd:digits="2">505.0</pbd:unit-price>
    </pbd:item>
    <pbd:item>
      <pbd:qty pbd:digits="0">601.0</pbd:qty>
      <pbd:description>
        <pbd:title>F2</pbd:title>
      </pbd:description>
      <pbd:unit-price pbd:digits="6">603.0</pbd:unit-price>
      <pbd:vat-rate>604.0</pbd:vat-rate>
    </pbd:item>
    <pbd:item pbd:holdback-rate="7.0" pbd:holdback-on-vat="no">
      <pbd:qty pbd:digits="0">701.0</pbd:qty>
      <pbd:description>
        <pbd:title>G2</pbd:title>
      </pbd:description>
      <pbd:unit-price pbd:digits="2">703.0</pbd:unit-price>
      <pbd:vat-rate>704.0</pbd:vat-rate>
    </pbd:item>
    <pbd:item pbd:holdback-rate="8.0" pbd:holdback-on-vat="yes">
      <pbd:qty pbd:digits="0">801.0</pbd:qty>
      <pbd:description>
        <pbd:title>H2</pbd:title>
      </pbd:description>
      <pbd:unit-price pbd:digits="2">803.0</pbd:unit-price>
      <pbd:vat-rate>804.0</pbd:vat-rate>
    </pbd:item>
  </pbd:items-list>
</pbd:accounting-document>
"""

class AccDocFormat10WriterProFormaTest(TestCase, FormatWriterGenericTest):
    def setUp(self):
        """
        Called before each test of this class.
        """
        self.accdoc = ProForma(u"A1")
        self._fill_accdoc()
        self.accdoc.validity_date = u"Z0"

    def test_write_xml_accdoc(self):
        xml_elt = write_accdoc_1_0_xml(self.accdoc)
        self.assertMultiLineEqual(etree.tostring(xml_elt, pretty_print=True),
                              EXPECTED)


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-
"""
``pybill.lib.config.entities`` module defines the class used to encapsulate the
configuration data.

.. autoclass:: ConfigData
   :members:
   :undoc-members:
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"

from pybill.lib.entities.addresses import OrganisationAddress


class ConfigData:
    """
    Class holding the configuration data.

    This configuration data is used to write the accounting documents in a PDF
    format. It mainly contains the company address and the localised words for
    the accounting documents.

    .. attribute:: reference

        Attribute defining the reference of the configuration that is used to
        store it in the configuration register. This reference is unique among
        configurations and matches the name of the XML file where the
        configuration is stored. It is also used to link the accounting XML
        documents with this configuration through the PyBill processing
        instruction.

       type:class:`unicode`

    .. attribute:: name

       Attribute defining the name of the configuration. The name can be used
       when displaying a list of the various configurations and is more explicit
       that the reference.

       type: :class:`unicode`

    .. attribute:: company_logo

       Attribute defining the name of the file containing the company logo (an
       image).  This logo will be displayed on every PDF document.

       type: :class:`str`

    .. attribute:: company_address

       Attribute containing the specific object that describes the company
       address.

       type: :class:`~pybill.lib.entities.addresses.OrganisationAddress`

    .. attribute:: footer_lines

       Attribute containing the list of the lines to be written in the footer
       of all the PDF documents.

       type: list of :class:`unicode`

    .. attribute:: bank_data_lines

       Attribute containing the list of the lines that describe the bank
       data. This information will be written on the PDF documents that require
       a payment from the client (bill, etc.)

       type: list of :class:`unicode`

    .. attribute:: agreement_intro_lines

       Attribute containing the list of the agreement lines to be written on
       the pro-forma PDF document (e.g. u"sign here").

       type: list of :class:`unicode`

    .. attribute:: number_separators

       Attribute containing a dictionnary describing the characters used to
       format the numbers in the PDF output. There are three separators
       corresponding to the three keys of the dictionary:

       - ``u"sign"`` contains the character separating the sign and the
         non-signed part of the number,

       - ``u"thousands"`` contains the character separating the thousands and
         the hundreds, the millions and the hundreds of thounsands, and so on,

       - ``u"digits"`` contains the character separating the integer part of the
         number and the digits.

       An example of ``number_separators`` dictionnary is ``{u"sign": u"",
       u"thousands": u" ", \ u"digits": u"."}```

       type: dictionary of :class:`unicode` indexed with :class:`unicode` keys

    .. attribute:: local_phrases

       Attribute containing a dictionary whose keys are the keywords used in
       PyBill and whose values are the localized phrases corresponding to these
       keywords (e.g. u"bill": u"Facture").
                         
       type: dictionary of :class:`unicode` indexed with :class:`unicode` keys

    .. automethod:: __init__
    """
    def __init__(self):
        """
        Intializes a new object.

        This object is empty and will be filled when reading an XML
        configuration file (cf.
        :class:`~pybill.lib.entities.config.xmlreaders.ConfigXMLReader`\ ).
        """
        self.reference = None
        self.name = u""
        self.company_logo = u""    
        self.company_address = OrganisationAddress()
        self.footer_lines = []
        self.bank_data_lines = []
        self.agreement_intro_lines = []
        self.number_separators = {}
        self.local_phrases = {}

    def get_local_phrase(self, keyword):
        """
        Gets the local phrase corresponding to the keyword ``keyword``.

        If the keyword is not defined in the configuration data, returns the
        keyword itself between two ``#`` characters (and doesn't raise any
        exception).

        :parameter keyword: keyword that must be localized. Can be for example
                            u"bill").
        :type keyword: :class:`unicode`

        :returns: The localized phrase associated to the given keyword in the
                  configuration data.
        :rtype: :class:`unicode`
        """
        if self.local_phrases.has_key(keyword):
            return self.local_phrases[keyword]
        else:
            return u"## %s ##" % keyword



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main

from pybill.lib.entities.accounting_docs import ClaimForm

from generic_test_accounting_docs import AbstractAccDocTest


class ClaimFormTest(TestCase, AbstractAccDocTest):
    """
    Tests on the claim form objects
    """
    def setUp(self):
        """
        Called before each test from this class
        """
        self.acc_doc = ClaimForm("ID1")


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-
"""
``pybill.lib.pdfwriters.writers`` module defines the public class that can
generate a PDF file for the accounting documents.

The PDF generation uses the platypus machinery of ReportLab and is, of course,
based on PyBill flow generators and templates.

.. autoclass:: PDFWriter
   :members:
   :undoc-members:
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"

from traceback import format_exc

from pybill.lib.errors import PyBillProcessingException

from pybill.lib.pdfwriters.templates import AccDocDocumentTemplate
from pybill.lib.pdfwriters.flowgenerators import (BillFlowGenerator,
                                                  ClaimFormFlowGenerator,
                                                  DownpaymentFlowGenerator,
                                                  ProFormaFlowGenerator,
                                                  DebitFlowGenerator)


class PDFWriter:
    """
    Class writing the PDF for all kinds of accounting documents.

    This class uses flow generators that are specific to the kind of accounting
    document to be rendered in PDF (bill, claim form, debit, downpayment,
    pro-forma). In order to prevent the instantiation of several generators of
    the same class, one flow generator of each class is kept in an attribute.

    .. attribute:: flow_generators

        Attribute containing a dictionary of flow generator objects indexed with
        the name of the class of the accounting documents they can transform in
        PDF. This dictionary contains all the specialized flow generators that
        can be used by this writer.

        type: Dictionary of :class:`~pybill.lib.pdfwriters.flowgenerators.abstract.AbstractFlowGenerator`
        objects indexed with unicode keys

    .. automethod:: __init__
    """

    def __init__(self):
        """
        Initializes a `PDFWriter` object and fills the ``flow_generators``
        attributes with the correct generator objects
        """
        self.flow_generators = {
            'Bill': BillFlowGenerator(),
            'ClaimForm': ClaimFormFlowGenerator(),
            'ProForma': ProFormaFlowGenerator(),
            'Downpayment': DownpaymentFlowGenerator(),
            'Debit' : DebitFlowGenerator(),
            }

    def write(self, acc_doc, pdf_file):
        """
        Writes a PDF file for an accounting document.

        :parameter acc_doc:
            Accounting document that must be transformed in PDF. This document
            contains the configuration object that will be used during this
            transform (a configuration object contains numerous useful pieces of
            data).

        :type acc_doc:
            :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`

        :parameter pdf_file:
            Name of the file where the PDF will be written in.
        :type pdf_file: unicode
        """
        try:
            flow_gen = self.flow_generators[acc_doc.__class__.__name__]
        except KeyError:
            raise PyBillProcessingException("Unexpected class of accounting "
                                            "document (%s) that can't be "
                                            "rendered in PDF."
                                            % acc_doc.__class__.__name__)
        try:
            # Opens a new PDF document from the PyBill-specific template
            pdf_doc = AccDocDocumentTemplate(pdf_file, acc_doc)
            # Fills it with the content returned by the flow generator
            pdf_doc.build(flow_gen.generate_flow(acc_doc))
        except IOError, exc:
            raise PyBillProcessingException("Unable to write PDF result in "
                                   "%s.\n%s" % (pdf_file, format_exc(limit=1)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
PyBill program generating the examples

Generates the PDF rendering for the example XML files. XML input files and
PDF output files are saved in a subdirectory.
"""

from optparse import OptionParser
import logging
import os.path as osp
from os import listdir

from pybill.lib.controller import PyBillController


logging.basicConfig(level=logging.DEBUG)


EXAMPLES_DIR = osp.abspath(osp.dirname(__file__))
CONFIGS_DIR = "configs"
INPUTS_DIR = "inputs"
RESULTS_DIR = "results"


def run(config_file=None) :
    """
    Function generating the examples. XML input files and PDF output files
    are saved in a subdirectory.

    :param config_file: Filename of a config file that will override config
                        references specified in the examples files.
    """
    logging.info("*** Generating examples in `%s` subdirectory"
                 % RESULTS_DIR)
    controller = PyBillController()
    # Hacking config register to read the config files in example config
    # subdirectory.
    controller.config_register.__class__.USER_CONFIG_DIR \
                                          = osp.join(EXAMPLES_DIR, CONFIGS_DIR)
    # Gets the example XML files to process.
    input_files = [osp.join(EXAMPLES_DIR, INPUTS_DIR, f)
                   for f in listdir(osp.join(EXAMPLES_DIR, INPUTS_DIR))
                   if osp.splitext(f)[1] == ".xml"]
    logging.info("    - Found %d example files in `%s` subdirectory"
                 % (len(input_files), INPUTS_DIR))
    if config_file is not None:
        logging.info("    - Overriding specified configurations with  %s config"
                     " file" % config_file)
    # Reads the accounting entities from the XML files.
    logging.info("    - Reading example files")
    read_entities = controller.build_entities_from_files(input_files,
                                                      cfg_filename=config_file)
    # For each accounting entity, changes the company logo filename in the
    # config object (puts an absolute path) and builds the pdf result filename.
    entities = []
    for accdoc, filename in read_entities:
        accdoc.cfg_data.company_logo = osp.join(EXAMPLES_DIR, CONFIGS_DIR,
                                                accdoc.cfg_data.company_logo)
        pdf_filename = "%s%s%s" % (osp.join(EXAMPLES_DIR, RESULTS_DIR,
                                            osp.split(filename)[1]),
                                   osp.extsep,
                                   "pdf")
        entities.append((accdoc, pdf_filename))
    # Generating PDF for the entities.
    logging.info("    - Generating PDF from example files")
    controller.write_pdf_to_files(entities)
    logging.info("    - Generated %d PDF files in `%s` subdirectory"
                 % (len(entities), RESULTS_DIR))
    logging.info("*** End of generation")


if __name__ == '__main__' :
    desc = ("Generates the PDFs of the examples of accounting documents with "
            "PyBill tool.")
    parser = OptionParser(description=desc)
    parser.add_option("-c", "--config",
                      action="store", type="string", dest="config_file",
                      default=None, metavar="CONFIG",
                      help=("override configs specified in examples documents "
                            "with CONFIG file"))
    (options, args) = parser.parse_args()
    run(options.config_file)



# -*- coding: utf-8 -*-
"""
``pybill.lib.pdfwriters.utils`` module defines two specific flowable elements
for :mod:`reportlab.platypus` and useful functions.

The flowable elements are the elements that can be inserted in the PDF by the
platypus machinery of Reportlab. These specific elements allow the computation
of a carry-forward for each page.

.. autoclass:: NumberParagraph
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: TotalizableTable
   :members:
   :undoc-members:
   :show-inheritance:

The functions define in this module concern the rendering of texts and numbers,
and the conversion to the specific flowable elements defined above.

.. autofunction:: escape_text

.. autofunction:: format_number

.. autofunction:: turn_Table_in_TotalizableTable
"""
__docformat__ = "restructuredtext en"

from xml.sax.saxutils import escape

from reportlab.platypus import Paragraph
from reportlab.platypus import Table

from pybill.lib import ACCURACY


def escape_text(text, normalize_spaces=False):
    """
    Escapes the text to be written in the PDF document.
    
    This function must be called when the text is in a platypus ``Paragraph``
    as some characters must be escaped (e.g. & and <).
    
    .. warning::

       Do not call this function after the text has been tagged as the
       tags will be escaped (e.g. ``<b>hello</b>`` will be turned in
       ``&lt;b>Hello&lt;/b>``\ ).
    
    :parameter text: Text to be escaped.
    :type text: :class:`unicode`
    
    :parameter normalize_spaces:
        When set to ``True``, the function normalizes the whitespace characters
        (no space at begining and end, only one space between the words).
    :type normalize_spaces: :class:`bool`
    
    :returns: The normalized text.
    :rtype: :class:`unicode`
    """
    if normalize_spaces:
        string = u" ".join(text.split())
    else :
        string = text
    return escape(string)


def format_number(number, digits=ACCURACY, separators=None):
    """
    Formats a number to be written in the PDF document.

    This function uses separators defined in the ``separators`` dictionnary
    (separating characters between the sign and the number, between the
    thousands and between the integer and the digits).
    
    :parameter number: Float number to be formatted.
    :type number: :class:`float`
    
    :parameter digits: Number of digits to be displayed (the float number is
                       rounded if it has more digits). 
    :type digits: :class:`int`

    :parameter separators:

        Dictionary containing the separators used to format a number. There are
        three separators:

        - `sign` that is inserted between the sign and the non-signed number,

        - `thousands` that is inserted between the thousands and the hundreds,
          between the millions and the hundreds of thousands, and so on,

        - `digits` that is inserted between the integer and the digits.

        When not specified, default separators are used.  They are:
        ``{u'sign': u'', u'thousands': u'',\ u'digits': u'.'}``.

    :type separators: dictionary of :class:`unicode` indexed with :class:`unicode` keys
    
    :returns: The float number formatted in a string.
    :rtype: :class:`unicode`
    """
    result = u""
    # Setting default values in the ``separators`` dictionnary.
    if separators is None:
        separators = {}
    separators.setdefault(u'sign', u'')
    separators.setdefault(u'thousands', u'')
    separators.setdefault(u'digits', u'.')
    # If number is None, returns u""
    if number is None:
        return u""
    # Deals with the sign.
    if number < 0:
        number = -number
        result += u"-%s" % separators[u'sign']
    # Deals with the integer part (cuts it three digits by three digits)
    number = round(number, digits)
    if digits > 0:
        num_int = int(number)
        num_dgts = number - num_int
    else:
        num_int = int(number)
        num_dgts = 0.0
    if num_int == 0:
        num_str = u"0"
    else:
        num_str = u""
        while num_int > 0:
            units = num_int % 1000
            num_int /= 1000
            if num_int > 0:
                num_str = u"%s%03d%s" % (separators[u'thousands'],
                                         units, num_str)
            else:
                num_str = u"%d%s" % (units, num_str)
    result += num_str
    # Deals with the digits
    if digits > 0:
        str_mask = u"%%.%df" % digits
        rounded_str = str_mask % num_dgts
        result += u"%s%s" % (separators[u'digits'], rounded_str.split(".")[1])
    # And finally returns the result!
    return result


def turn_Table_in_TotalizableTable(table):
    """
    Builds a PyBill-specific :class:`TotalizableTable` object from a
    platypus standard :class:`~reportlab.platypus.Table` object.

    The two objects have the same internal content.
    
    :parameter table: Platypus table object to be turned in a PyBill-specific
                      :class:`TotalizableTable` object.
    :type table: :class:`reportlab.platypus.Table`
    
    :returns: Pybill-specific ``TotalizableTable`` object identical to the
              Platypus table.
    :rtype: :class:`TotalizableTable`
    """
    tot_table = TotalizableTable(table._cellvalues,
                                  table._colWidths, table._argH,
                                  repeatRows=table.repeatRows,
                                  repeatCols=table.repeatCols,
                                  splitByRow=table.splitByRow)
    # Copies the style instructions
    tot_table._cellStyles = table._cellStyles[:]
    tot_table._linecmds = table._linecmds[:]
    tot_table._bkgrndcmds = table._bkgrndcmds[:]
    tot_table._spanCmds = table._spanCmds[:]
    return tot_table


class NumberParagraph(Paragraph):
    """
    PyBill-specific paragraph that only contains a number.

    The ``NumberParagraph`` objects will be used by PyBill-specific
    :class:`TotalizableTable` class to compute a sum over the rows of the table,
    and be able to compute a carry-forward at the end of each page.

    Basically a ``NumberParagraph`` is identical to a platypus
    :class:`~reportlab.platypus.Paragraph` but contains a float number (see
    ``number`` attribute). This number will be used to compute the sum in a
    ``TotalizableTable``.

    .. attribute:: number

       Attribute containing the number displayed in the paragraph. It is this
       number that will be totalized in the ``TotalizableTable``.

       type: :class:`float`

    .. automethod:: __init__
    """
    def __init__(self, number, text, style,
                 bulletText = None, frags=None, caseSensitive=1):
        """
        Initializes a ``NumberParagraph``.

        Most of the parameters are identical to those of the platypus
        :class:`~reportlab.platypus.Paragraph` initializer (see this base
        class for details on the parameters).
        
        :parameter number: The number contained in the paragraph. This number
                           will be used by the ``TotalizableTable`` to compute
                           the total over its rows.

                           This parameter is specific to the
                           ``NumberParagraph`` class.
        :type number: :class:`float`
        
        :parameter text: The text contained in the paragraph (see ``Paragraph``
                         for a full description of the available markup).
        :type text: :class:`unicode`
        
        :parameter style: The style of the paragraph.
        :type style: :class:`reportlab.lib.styles.ParagraphStyle`
        """
        self.number = number
        Paragraph.__init__(self, text, style, bulletText = bulletText,
                           frags=frags, caseSensitive=caseSensitive)


class TotalizableTable(Table):
    """
    PyBill-specific table that can compute a total over its cells.

    The total is computed by summing all the the :class:`NumberParagraph`
    objects contained in the table cells.
    
    Usually there is only one ``NumberParagraph`` per row and thus the
    ``TotalizableTable`` can compute the sum of an amount over the rows.

    Please note that only the cells of the table that contain a single
    ``NumberParagraph`` object will be taken into account for computing the
    sum.

    .. automethod:: _splitRows
    """
    def __repr__(self):
        """
        Returns a string representation of the object.

        :returns: String representation of the object.
        :rtype: :class:`str`
        """
        return "Totalizable"+Table.__repr__(self)

    def _splitRows(self, availHeight):
        """
        Splits the table into two tables. The first table height is maximal
        but less than availHeight.
        
        :parameter availHeight: Available height in the PDF frame.
        :type availHeight: :class:`float`
        
        :returns: A list of ``TotalizableTable`` objects containing all the
                  rows of the table that have been split into several objects.
                  This list can be empty, have a single element or have
                  several elements.
        :rtype: list of :class:`TotalizableTable`
        """
        # Ask the ``_splitRows`` method of the base class to split the table.
        result = Table._splitRows(self, availHeight)
        # As the result contains ``Table`` instances and not
        # ``TotalizableTable`` instances, turns the result elements into
        # ``TotalizableTable`` objects.
        return [turn_Table_in_TotalizableTable(elt) for elt in result]

    def compute_total(self):
        """
        Computes the sum, over the cells of the table, of all the numbers
        contained in :class:`NumberParagraph` objects.
        
        Usually there is only one ``NumberParagraph`` per row, thus the
        ``compute_total`` method returns the sum over the rows of the table.

        Please note that only the cells of the table that contain a single
        ``NumberParagraph`` object will be taken into account for computing
        the sum.

        :returns: the sum of all the numbers contained in the
                  ``NumberParagraph`` objetcs inside the cells of the table.
        :rtype: :class:`float`
        """
        return sum([cell.number for line in self._cellvalues for cell in line
                    if cell.__class__ is NumberParagraph])



# -*- coding: utf-8 -*-
"""
``pybill.lib.xmlwriters.utils`` module defines constants and common functions
used to write accounting entities in XML.

These functions mainly concern opening an XML file and writing an address object
into an XML element with a DocBook-like structure.

This module defines a constant that will be widely used in PyBill.

.. autodata:: ENCODING

It also defines the following utilitary functions.

.. autofunction:: open_xml_outfile

.. autofunction::  build_address_xml_element

.. autofunction:: write_address_xml_elements
"""
__docformat__ = "restructuredtext en"

from traceback import format_exc
from lxml import etree

from pybill.lib.errors import PyBillWritingException
from pybill.lib import PBD_XMLNS


ENCODING = u"UTF-8"
"""
Constant specifying the default encoding used for the XML files written in 
this package.

type: :class:`unicode`
"""


def open_xml_outfile(xml_file, encoding=ENCODING):
    """
    Opens an XML file in write mode, and writes the XML declaration.

    :parameter xml_file: Name of the XML file. Also works if this parameter
                         is an opened file.
    :type xml_file: :class:`unicode`

    :parameter encoding: Encoding used to write the XML file.
    :type encoding: :class:`str`

    :returns: opened file in write mode where the XML declaration has already
              been written.
    :rtype: :class:`file`
    """
    if type(xml_file) in (str, unicode):
        # xml_file is a string, thus a real filename, we try to open the file.
        try:
            out = open(xml_file, "w")
        except IOError, exc:
            raise PyBillWritingException("Can't open the '%s' file for writing "
                                   "data\n%s" % (xml_file, format_exc(limit=1)))
    else:
        # xml_file is not a string, we suppose it is an already opened file.
        out = xml_file
    # Writes the XML declaration in the file
    try:
        out.write('<?xml version="1.0" encoding="%s"?>\n\n' % encoding)
    except IOError, exc:
        raise PyBillWritingException("Can't write data in '%s' file\n%s"
                                     % (xml_file, format_exc(limit=1)))
    return out


def build_address_xml_element(address, xml_ns=u""):
    """
    Transforms the address object in an XML element whose structure is inspired
    by DocBook.

    This type of XML element is used in several PyBill formats (\
    `PyBill Document 0.X` or `PyBill Document 1.0`\ ). The XML element is built
    thanks to :mod:`lxml` library.

    :parameter address: Address object to be turned into an XML element.
    :type address: :class:`~pybill.lib.entities.addresses.PersonAddress`

    :parameter xml_ns: XML namespace where the address XML elements will be
                       defined in. The default value ``u""`` corresponds to
                       an empty namespace.
    :type xml_ns: :class:`unicode`
    
    :returns: An XML element built with `lxml` library that contains the
              information from the ``address`` object.
    :rtype: `:class:lxml.etree.Element`
    """
    adr_elt = etree.Element(u"{%s}address" % xml_ns)
    if address.honorific.strip() != u"":
        hnr_elt = etree.SubElement(adr_elt, u"{%s}honorific" % xml_ns)
        hnr_elt.text = address.honorific
    if address.firstname.strip() != u"":
        fnm_elt = etree.SubElement(adr_elt, u"{%s}firstname" % xml_ns)
        fnm_elt.text = address.firstname
    if address.other_name.strip() != u"":
        onm_elt = etree.SubElement(adr_elt, u"{%s}othername" % xml_ns)
        onm_elt.text = address.other_name
    if address.surname.strip() != u"":
        snm_elt = etree.SubElement(adr_elt, u"{%s}surname" % xml_ns)
        snm_elt.text = address.surname
    if address.lineage.strip() != u"":
        lng_elt = etree.SubElement(adr_elt, u"{%s}lineage" % xml_ns)
        lng_elt.text = address.lineage
    write_address_xml_elements(address, adr_elt, xml_ns)
    if address.organisation is not None:
        org = address.organisation
        aff_elt = etree.SubElement(adr_elt, u"{%s}affiliation" % xml_ns)
        ogn_elt = etree.SubElement(aff_elt, u"{%s}orgname" % xml_ns)
        ogn_elt.text = org.name
        for odiv in [d for d in org.divisions if d.strip() != u""]:
            odv_elt = etree.SubElement(aff_elt, u"{%s}orgdiv" % xml_ns)
            odv_elt.text = odiv
        for jbtit in [j for j in org.job_titles if j.strip() != u""]:
            jbt_elt = etree.SubElement(aff_elt, u"{%s}jobtitle" % xml_ns)
            jbt_elt.text = jbtit
        org_adr_elt = etree.Element(u"{%s}address" % xml_ns)
        write_address_xml_elements(org, org_adr_elt, xml_ns)
        if org_adr_elt.getchildren() != []:
            aff_elt.append(org_adr_elt)
    return adr_elt


def write_address_xml_elements(address, xml_elt, xml_ns=u""):
    """
    Writes the elements of the ``address`` object inside the ``xml_elt``
    XML element, using a structure inspired by DocBook.

    This function can be used to dump the elements of a ``PersonAddress`` or
    an ``OrganisationAddress``. The dumped elements are ``street``,
    ``postcode``, etc. that are common to both addresses.
    
    :parameter address: Address object whose elements must be written
                        in XML into the XML element. This object is either
                        a ``PersonAddress`` or an ``OrganisationAddress``.
    :type address: :class:`~pybill.lib.entities.addresses.Address`

    :parameter xml_elt: XML element in which the XML elements representing
                        the address elements will be written. This XML
                        element is built with the `lxml` library and will be
                        modified by this function (addition of new child
                        elements).
    :type xml_elt: :class:`lxml.etree.Element`

    :parameter xml_ns: XML namespace where the address XML elements will be
                       defined in. The default value ``u""`` corresponds to
                       an empty namespace.
    :type xml_ns: :class:`unicode`
    """
    for street in [s for s in address.streets if s.strip() != u""]:
        str_elt = etree.SubElement(xml_elt, u"{%s}street" % xml_ns)
        str_elt.text = street
    if address.post_box.strip() != u"":
        pbx_elt = etree.SubElement(xml_elt, u"{%s}pob" % xml_ns)
        pbx_elt.text = address.post_box
    if address.post_code.strip() != u"":
        pcd_elt = etree.SubElement(xml_elt, u"{%s}postcode" % xml_ns)
        pcd_elt.text = address.post_code
    if address.city.strip() != u"":
        cty_elt = etree.SubElement(xml_elt, u"{%s}city" % xml_ns)
        cty_elt.text = address.city
    if address.state.strip() != u"":
        sta_elt = etree.SubElement(xml_elt, u"{%s}state" % xml_ns)
        sta_elt.text = address.state
    if address.country.strip() != u"":
        cnt_elt = etree.SubElement(xml_elt, u"{%s}country" % xml_ns)
        cnt_elt.text = address.country
    if address.phone.strip() != u"":
        phn_elt = etree.SubElement(xml_elt, u"{%s}phone" % xml_ns)
        phn_elt.text = address.phone
    if address.fax.strip() != u"":
        fax_elt = etree.SubElement(xml_elt, u"{%s}fax" % xml_ns)
        fax_elt.text = address.fax
    if address.web.strip() != u"":
        web_elt = etree.SubElement(xml_elt, u"{%s}web" % xml_ns)
        web_elt.text = address.web 
    if address.email.strip() != u"":
        eml_elt = etree.SubElement(xml_elt, u"{%s}email" % xml_ns)
        eml_elt.text = address.email



# -*- coding: utf-8 -*-
"""
``pybill.lib.pdfwriters.templates.page_templates`` module contains the
definition of the page template classes used in the PDF generation with
ReportLab `platypus`.

These classes are derived from platypus standard classes but are specific to
PyBill software. The page templates draw the static content specific to the
accounting documents.

.. autoclass:: AccDocPageTemplate
   :show-inheritance:

.. autoclass:: AccDocFirstPageTemplate
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"

from locale import getdefaultlocale
from traceback import format_exc
import logging

from reportlab.lib.units import cm
from reportlab.lib.colors import black

from reportlab.platypus import Paragraph, Spacer, Image, Frame
from reportlab.platypus import Table, TableStyle
from reportlab.platypus import PageTemplate, BaseDocTemplate

from pybill.lib.errors import PyBillReadingException, PyBillProcessingException
from pybill.lib.pdfwriters.utils import escape_text, format_number
from pybill.lib import ACCURACY
from pybill.lib.pdfwriters import (style_title, style_footer,
                                   style_normal_left, style_normal_right,
                                   style_little_left, style_little_right)


# Constant used to debug the result PDF (shows frame boundaries in the PDF).
SHOW_FRAMES_ON_PDF = False


class AccDocPageTemplate(PageTemplate):
    """
    Class defining the page template used to draw the other pages of accounting
    PDF documents (not the first one).

    This page template has a main frame that will be filled with the accounting
    table, and various other frames containing the logo, and the footer that are
    displayed identically on each page (static content).
    
    This page template can only be used with a document created from the
    :class:`~pybill.lib.pdfwriters.templates.doc_templates.AccDocDocumentTemplate`
    document template. It contains some methods for writing the carry forward
    computed in this document template.

    .. attribute:: acc_doc
    
       Attribute containing the accounting entity thate describes the actual
       document to be rendered in PDF.
                   
       type: :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`

    .. automethod:: __init__

    .. automethod:: _create_main_frame

    .. automethod:: beforeDrawPage

    .. automethod:: afterDrawPage

    .. automethod:: _draw_static_content

    .. automethod:: _draw_top_carry_forward

    .. automethod:: _draw_bottom_carry_forward
    """
    def __init__(self, acc_doc, template_id=""):
        """
        Initializes an ``AccDocPageTemplate`` object.

        :parameter acc_doc: Accounting entity containing the actual document
                            to be rendered in PDF.
        :type acc_doc:
            :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`

        :parameter template_id: identifier (name) of the page template.
        :type template_id: :class:`str`
        """
        self.acc_doc = acc_doc
        # Defines the main frame in which the accounting table will be
        # displayed.
        main_f = self._create_main_frame()
        # Calls the base class initializer.
        PageTemplate.__init__(self, id=template_id, frames=[main_f])
    
    def _create_main_frame(self):
        """
        Creates the main frame of the page template, in which the flow,
        generated from the accounting entity, will be displayed.
        
        :returns: The frame in which the PDF flow will be rendered.
        :rtype: :class:`reportlab.platypus.Frame`
        """
        main_f = Frame(1*cm, 4.5*cm, 19*cm, 18.3*cm,
                       leftPadding=0*cm, bottomPadding=0*cm,
                       rightPadding=0*cm, topPadding=0*cm, 
                       showBoundary=SHOW_FRAMES_ON_PDF)
        return main_f

    def beforeDrawPage(self, canvas, document):
        """
        Overrides the method called before drawing a new page of the PDF
        document.

        This method draws various static data on the page (logo, company
        address, footer, page number, carry forward from previous page).
        
        :parameter canvas: Canvas of the page where the content can be written.
        :type canvas: :class:`reportlab.pdfgen.canvas.Canvas`
        
        :parameter document:
            PDF Document created from a template. This document contains various
            useful information such as the page number or the carry forward.
        :type document:
            :class:`~pybill.lib.pdfwriters.templates.doc_templates.AccDocDocumentTemplate`
        """
        # Saves canvas state.
        canvas.saveState()
        self._draw_static_content(canvas, document)
        if document.display_carry_forward:
            self._draw_top_carry_forward(canvas, document)
        # Puts back the canvas in its previous state.
        canvas.restoreState()

    def afterDrawPage(self, canvas, document):
        """
        Overrides the method called after drawing a page of the PDF document.

        This method draws the carry forward to be brought forward to the next
        page.

        :parameter canvas: Canvas of the page where content can be written.
        :type canvas: `reportlab.pdfgen.canvas.Canvas`
        
        :parameter canvas: Canvas of the page where the content can be written.
        :type canvas: :class:`reportlab.pdfgen.canvas.Canvas`
        
        :parameter document:
            PDF Document created from a template. This document contains various
            useful information such as the page number or the carry forward.
        :type document:
            :class:`~pybill.lib.pdfwriters.templates.doc_templates.AccDocDocumentTemplate`
        """
        # Saves canvas state.
        canvas.saveState()
        if document.display_carry_forward:
            self._draw_bottom_carry_forward(canvas, document)
        # Puts back the canvas in its previous state.
        canvas.restoreState()

    def _draw_static_content(self, canvas, document):
        """
        Draws on the PDF page the static content that is displayed on all
        pages, i.e. logo, company address, footer, page number.

        :parameter canvas: Canvas of the page where the content can be written.
        :type canvas: :class:`reportlab.pdfgen.canvas.Canvas`
        
        :parameter document:
            PDF Document created from a template. This document contains various
            useful information such as the page number.
        :type document:
            :class:`~pybill.lib.pdfwriters.templates.doc_templates.AccDocDocumentTemplate`
        """
        cfg_data = self.acc_doc.cfg_data
        # Draws the logo.
        if cfg_data.company_logo: 
            # The logo is not drawn through a frame because images as flowables
            # cannot be aligned on the left (they are always centered).
            # The logo filename is converted to a string (no unicode!)
            if cfg_data.company_logo.__class__ is unicode:
                logo_filename = cfg_data.company_logo.encode(
                                                        getdefaultlocale()[1])
            else:
                logo_filename = cfg_data.company_logo
            try:
                img = Image(logo_filename)
            except IOError, exc:
                raise PyBillReadingException("Can't open logo file '%s'.\n%s"
                                        % (logo_filename, format_exc(limit=1)))
            img_height  = 1.8*cm
            img_width = img_height * img.drawWidth / img.drawHeight
            canvas.drawInlineImage(logo_filename,
                                   1*cm, 26.9*cm, img_width, img_height)
        else:
            # If there is no logo, writes the company name.
            if cfg_data.company_address.name:
                name_f = Frame(1*cm, 26.9*cm, 8.5*cm, 1.8*cm,
                               leftPadding=0*cm, bottomPadding=0*cm,
                               rightPadding=0*cm, topPadding=0*cm, 
                               showBoundary=SHOW_FRAMES_ON_PDF)
                f_content = [Paragraph(u"<b>%s</b>"
                                  % escape_text(cfg_data.company_address.name),
                                                style_title)]
                name_f.addFromList(f_content, canvas)
                if len(f_content) > 0:
                    text = (u"The name of the sending company ('%s') is too "
                            u"long and can't be displayed in the PDF."
                             % cfg_data.company_address.name)
                    raise PyBillProcessingException(
                                     text.encode("ascii", "xmlcharrefreplace"))
        # Draws the company address in a dedicated frame.
        company_f = Frame(1*cm, 22.9*cm, 8.5*cm, 3.9*cm,
                          leftPadding=0*cm, bottomPadding=0*cm,
                          rightPadding=0*cm, topPadding=0*cm, 
                          showBoundary=SHOW_FRAMES_ON_PDF)
        f_content = []
        for line in cfg_data.company_address.get_postal_address_lines():
            f_content.append(Paragraph(escape_text(line),
                                       style_normal_left))
        if cfg_data.company_address.phone != u"":
            text = u"%s%s %s" % (cfg_data.get_local_phrase(u'phone-kw'),
                                 cfg_data.get_local_phrase(u'colon'),
                                 cfg_data.company_address.phone)
            f_content.append(Paragraph(escape_text(text), style_little_left))
        if cfg_data.company_address.fax != u"":
            text = u"%s%s %s" % (cfg_data.get_local_phrase(u'fax-kw'),
                                 cfg_data.get_local_phrase(u'colon'),
                                 cfg_data.company_address.fax)
            f_content.append(Paragraph(escape_text(text), style_little_left))
        if cfg_data.company_address.web != u"":
            text = u"%s%s %s" % (cfg_data.get_local_phrase(u'web-kw'),
                                 cfg_data.get_local_phrase(u'colon'),
                                 cfg_data.company_address.web)
            f_content.append(Paragraph(escape_text(text), style_little_left))
        if cfg_data.company_address.email != u"":
            text = u"%s%s %s" % (cfg_data.get_local_phrase(u'email-kw'),
                                 cfg_data.get_local_phrase(u'colon'),
                                 cfg_data.company_address.email)
            f_content.append(Paragraph(escape_text(text), style_little_left))
        company_f.addFromList(f_content, canvas)
        if len(f_content) > 0:
            text = ("The address of the sending company is too long and can't "
                    "be entirely displayed in the PDF. The following lines "
                    "are not displayed:\n")
            for para in f_content:
                text += "  %s\n" % para.text
            raise PyBillProcessingException(
                                     text.encode("ascii", "xmlcharrefreplace"))
        # Draws the footer data (actually the company address) in a
        # dedicated frame.
        footer_f = Frame(3*cm, 1*cm, 15*cm, 2*cm,
                         leftPadding=0*cm, bottomPadding=0*cm,
                         rightPadding=0*cm, topPadding=0*cm, 
                         showBoundary=SHOW_FRAMES_ON_PDF)
        f_content = []
        for line in [l for l in cfg_data.footer_lines if l != u""]:
            string = u"<i>%s</i>" % escape_text(line)
            f_content.append(Paragraph(string, style_footer))
        footer_f.addFromList(f_content, canvas)
        if len(f_content) > 0:
            text = ("The footer is too long and can't be entirely displayed "
                    "in the PDF. The following lines are not displayed:\n")
            for para in f_content:
                text += "  %s\n" % para.text
            raise PyBillProcessingException(
                                     text.encode("ascii", "xmlcharrefreplace"))
        # Draws the header data (document reference and page number) in a
        # dedicated frame.
        pagenum_f = Frame(12.5*cm, 27.7*cm, 7.5*cm, 1*cm,
                          leftPadding=0*cm, bottomPadding=0*cm,
                          rightPadding=0*cm, topPadding=0*cm,
                          showBoundary=SHOW_FRAMES_ON_PDF)
        if self.acc_doc.doc_ref.strip() != u"":
            text = u"%s - " % self.acc_doc.doc_ref.strip()
        else:
            text = u""
        text += u"Page %d" % document.page
        f_content = [Paragraph(escape_text(text, True),
                               style_little_right)]
        pagenum_f.addFromList(f_content, canvas)
        if len(f_content) > 0:
            text = ("The document reference ('%s') is too long and can't be "
                    "displayed in the PDF (in page header)." 
                    % self.acc_doc.doc_ref)
            raise PyBillProcessingException(
                                     text.encode("ascii", "xmlcharrefreplace"))

    def _draw_top_carry_forward(self, canvas, document):
        """
        Draws on the PDF page the carry forward that is brought forward from
        the previous page.

        This carry forward is displayed at the top of the page just before the
        main frame.

        :parameter canvas: Canvas of the page where the content can be written.
        :type canvas: :class:`reportlab.pdfgen.canvas.Canvas`
        
        :parameter document:
            PDF Document created from a template. This document contains various
            useful information such as the carry forward.
        :type document:
            :class:`~pybill.lib.pdfwriters.templates.doc_templates.AccDocDocumentTemplate`
        """
        cfg_data = self.acc_doc.cfg_data
        c_fwd_f = Frame(12.5*cm, 23*cm, 7.5*cm, 0.8*cm,
                        leftPadding=0*cm, bottomPadding=0*cm,
                        rightPadding=0*cm, topPadding=0*cm,
                        showBoundary=SHOW_FRAMES_ON_PDF)
        # Builds the cells list for the carry forward table (1 line, 2 cols).
        cf_table_data = [[]]
        cf_table_data[0].append(Paragraph(
                      escape_text(cfg_data.get_local_phrase(u'carry-forward')),
                                          style_normal_left))
        cf_table_data[0].append(Paragraph(
                                format_number(document.carry_forward, ACCURACY,
                                              cfg_data.number_separators),
                                          style_normal_right))
        # Builds the carry forward table (with its style).
        cf_table = Table(cf_table_data, [4.5*cm, 3*cm], None)
        cf_table.setStyle(
            TableStyle([ ('ALIGN',     (0,0),( 0,-1), 'LEFT'),
                         ('ALIGN',     (1,0),( 1,-1), 'RIGHT'),
                         ('VALIGN',    (0,0),(-1,-1), 'TOP'),
                         ('INNERGRID', (0,0),(-1,-1), 0.25, black),
                         ('BOX',       (0,0),(-1,-1), 0.5, black),
                       ])
            )
        f_content = [cf_table]
        c_fwd_f.addFromList(f_content, canvas)
        if len(f_content) > 0:
            text = ("The carry forward can't be displayed in the PDF, on the "
                    "top of the page (lacking space).")
            raise PyBillProcessingException(
                                     text.encode("ascii", "xmlcharrefreplace"))

    def _draw_bottom_carry_forward(self, canvas, document):
        """
        Draws on the PDF page the carry forward that is brought forward to
        the next page.

        This carry forward is displayed at the bottom of the page just after the
        main frame.

        :parameter canvas: Canvas of the page where the content can be written.
        :type canvas: :class:`reportlab.pdfgen.canvas.Canvas`
        
        :parameter document:
            PDF Document created from a template. This document contains various
            useful information such as the page number.
        :type document:
            :class:`~pybill.lib.pdfwriters.templates.doc_templates.AccDocDocumentTemplate`
        """
        cfg_data = self.acc_doc.cfg_data
        c_fwd_f = Frame(12.5*cm, 3.5*cm, 7.5*cm, 0.8*cm,
                        leftPadding=0*cm, bottomPadding=0*cm,
                        rightPadding=0*cm, topPadding=0*cm,
                        showBoundary=SHOW_FRAMES_ON_PDF)
        # Builds the cells list for the carry forward table (1 line, 2 cols).
        cf_table_data = [[]]
        cf_table_data[0].append(Paragraph(
                   escape_text(cfg_data.get_local_phrase(u'to-bring-forward')),
                                          style_normal_left))
        cf_table_data[0].append(Paragraph(
                                format_number(document.carry_forward, ACCURACY,
                                              cfg_data.number_separators),
                                          style_normal_right))
        # Builds the carry forward table (with its style).
        cf_table = Table(cf_table_data, [4.5*cm, 3*cm], None)
        cf_table.setStyle(
            TableStyle([ ('ALIGN',     (0,0),( 0,-1), 'LEFT'),
                         ('ALIGN',     (1,0),( 1,-1), 'RIGHT'),
                         ('VALIGN',    (0,0),(-1,-1), 'TOP'),
                         ('INNERGRID', (0,0),(-1,-1),  0.25, black),
                         ('BOX',       (0,0),(-1,-1),  0.5, black),
                       ])
            )
        f_content = [cf_table]
        c_fwd_f.addFromList(f_content, canvas)
        if len(f_content) > 0:
            text = ("The carry forward can't be displayed in the PDF, on the "
                    "bottom of the page (lacking space).")
            raise PyBillProcessingException(
                                     text.encode("ascii", "xmlcharrefreplace"))


class AccDocFirstPageTemplate(AccDocPageTemplate):
    """
    Class defining the page template used to draw the first pages of accounting
    documents in PDF.

    This page template has a main frame that will be filled
    with the accounting table, and various other frames containing the logo,
    the date, the address, the footer that are displayed identically on each
    page (static content).
    
    This template is a specialization of
    :class:`~pybill.lib.pdfwriters.templates.page_templates.AccDocPageTemplate`.
    
    This page template can only be used with a document created from the
    :class:`~pybill.lib.pdfwriters.templates.doc_templates.AccDocDocumentTemplate`
    document template. It contains some methods for writing the carry forward
    computed in this document template.

    .. automethod:: __init__
        
    .. automethod:: _create_main_frame
        
    .. automethod:: beforeDrawPage
        
    .. automethod:: _draw_specific_static_content
    """
    def __init__(self, acc_doc, template_id=""):
        """
        Initializes an ``AccDocFirstPageTemplate`` object.

        :parameter acc_doc: Accounting entity containing the actual document
                            to be rendered in PDF.
        :type acc_doc:
            :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`

        :parameter template_id: identifier (name) of the page template.
        :type template_id: :class:`str`
        """
        AccDocPageTemplate.__init__(self, acc_doc, template_id=template_id)

    def _create_main_frame(self):
        """
        Creates the main frame of the page template, in which the flow,
        generated from the accounting entity, will be displayed.

        The main frame on the first page is littler than on the other pages as
        there is more static content on the first page (addresses, etc.)
        
        :returns: The frame in which the PDF flow will be rendered.
        :rtype: :class:`reportlab.platypus.Frame`
       """
        main_f = Frame(1*cm, 4.5*cm, 19*cm, 13.3*cm,
                       leftPadding=0*cm, bottomPadding=0*cm,
                       rightPadding=0*cm, topPadding=0*cm, 
                       showBoundary=SHOW_FRAMES_ON_PDF)
        return main_f

    def beforeDrawPage(self, canvas, document):
        """
        Overrides the method called before drawing a new page of the PDF
        document.

        This method draws various static data on the page (logo, company
        address, footer, page number, carry forward from previous page).
        
        :parameter canvas: Canvas of the page where the content can be written.
        :type canvas: :class:`reportlab.pdfgen.canvas.Canvas`
        
        :parameter document:
            PDF Document created from a template. This document contains various
            useful information such as the page number.
        :type document:
            :class:`~pybill.lib.pdfwriters.templates.doc_templates.AccDocDocumentTemplate`
        """
        # Saves canvas state.
        canvas.saveState()
        self._draw_static_content(canvas, document)
        self._draw_specific_static_content(canvas, document)
        # Puts back the canvas in its previous state.
        canvas.restoreState()

    def _draw_specific_static_content(self, canvas, document):
        """
        Draws on the PDF page some specific static content that is displayed
        on each page based on this template (in fact, only the first page).
        
        This specific static content is document metadata, receiver address,
        and date.
        
        :parameter canvas: Canvas of the page where the content can be written.
        :type canvas: :class:`reportlab.pdfgen.canvas.Canvas`
        
        :parameter document:
            PDF Document created from a template. This document contains various
            useful information such as the page number.
        :type document:
            :class:`~pybill.lib.pdfwriters.templates.doc_templates.AccDocDocumentTemplate`
        """
        cfg_data = self.acc_doc.cfg_data
        # Draws the metadata (document various references) in a dedicated
        # frame.
        metad_f = Frame(1*cm, 19.95*cm, 8.5*cm, 2.75*cm, 
                        leftPadding=0*cm, bottomPadding=0*cm,
                        rightPadding=0*cm, topPadding=0*cm, 
                        showBoundary=SHOW_FRAMES_ON_PDF)
        f_content = []
        if self.acc_doc.doc_ref.strip() != u"":
            text = u"%s%s %s" % (cfg_data.get_local_phrase(u'doc-ref-kw'),
                                 cfg_data.get_local_phrase(u'colon'),
                                 self.acc_doc.doc_ref.strip())
            f_content.append(Paragraph(escape_text(text, True),
                                       style_little_left))
        for kw, val in [(k, v) for k, v in self.acc_doc.other_infos.items()
                        if (k.strip() != u"" or v.strip() != u"")]:
            text = u"%s%s %s" % (kw.strip(),
                                 cfg_data.get_local_phrase('colon'),
                                 val.strip())
            f_content.append(Paragraph(escape_text(text, True),
                                       style_little_left))
        metad_f.addFromList(f_content, canvas)
        if len(f_content) > 0:
            text = ("There is too many metadata information in document. The "
                    "following items can't be displayed in the PDF:\n")
            for para in f_content:
                text += "  %s\n" % para.text
            logging.warning(text.encode("ascii", "xmlcharrefreplace"))
        # Draws the receiver company address in a dedicated frame.
        address_f = Frame(9.8*cm, 21.85*cm, 9.7*cm, 3.7*cm,
                          leftPadding=0*cm, bottomPadding=0*cm,
                          rightPadding=0*cm, topPadding=0*cm, 
                          showBoundary=SHOW_FRAMES_ON_PDF)
        f_content = []
        if self.acc_doc.receiver.organisation:
            text = self.acc_doc.receiver.organisation.name.strip()
            if text != u"":
                f_content.append(Paragraph(u"<b>%s</b>" 
                                           % escape_text(text, True),
                                           style_normal_left))
            for div in [d.strip() for d
                        in self.acc_doc.receiver.organisation.divisions
                        if d.strip() != u""]:
                f_content.append(Paragraph(escape_text(div, True),
                                           style_normal_left))
            for line in \
                 self.acc_doc.receiver.organisation.get_postal_address_lines():
                f_content.append(Paragraph(escape_text(line, True),
                                           style_normal_left))
        else:
            text = self.acc_doc.receiver.get_person_name()
            if text != u"":
                f_content.append(Paragraph(u"<b>%s</b>" 
                                           % escape_text(text, True),
                                           style_normal_left))
            for line in self.acc_doc.receiver.get_postal_address_lines():
                f_content.append(Paragraph(escape_text(line, True),
                                           style_normal_left))
        address_f.addFromList(f_content, canvas)
        if len(f_content) > 0:
            text = ("The receiver's address is too long and can't be entirely "
                    "displayed in the PDF. The following lines are not "
                    "displayed:\n")
            for para in f_content:
                text += "  %s\n" % para.text
            logging.warning(text.encode("ascii", "xmlcharrefreplace"))
        # Draws the sender, receiver, date and place in a dedicated frame.
        date_f = Frame(1*cm, 17.9*cm, 19*cm, 1.95*cm,
                       leftPadding=0*cm, bottomPadding=0*cm,
                       rightPadding=0*cm, topPadding=0*cm, 
                       showBoundary=SHOW_FRAMES_ON_PDF)
        f_content = []
        if self.acc_doc.receiver.get_person_name() != u"":
            text = u"%s%s <b>%s</b>"\
                 % (escape_text(cfg_data.get_local_phrase(u'receiver-kw')),
                    escape_text(cfg_data.get_local_phrase(u'colon')),
                    escape_text(self.acc_doc.receiver.get_person_name(), True))
            if self.acc_doc.receiver.organisation:
                for job_t in [j.strip() for j
                              in self.acc_doc.receiver.organisation.job_titles
                              if j.strip() != u""]:
                    text += u" - %s" % escape_text(job_t, True)
            f_content.append(Paragraph(text, style_little_left))
        if (self.acc_doc.sender is not None and 
            self.acc_doc.sender.get_person_name() != u""):
            text = u"%s%s <b>%s</b>" \
                   % (escape_text(cfg_data.get_local_phrase(u'sender-kw')),
                      escape_text(cfg_data.get_local_phrase(u'colon')),
                      escape_text(self.acc_doc.sender.get_person_name(), True))
            if self.acc_doc.sender.phone.strip() != u"":
                text += u" - %s" % escape_text(self.acc_doc.sender.phone, True)
            if self.acc_doc.sender.email.strip() != u"":
                text += u" - %s" % escape_text(self.acc_doc.sender.email, True)
            f_content.append(Paragraph(text, style_little_left))
        f_content.append(Spacer(1, 0.2*cm))
        text = u"%s, %s %s" % (self.acc_doc.place.strip(),
                               cfg_data.get_local_phrase(u'on-date'),
                               self.acc_doc.date)
        f_content.append(Paragraph(escape_text(text, True),
                                   style_normal_right))
        date_f.addFromList(f_content, canvas)
        if len(f_content) > 0:
            text = ("The date, the sender's name and the receiver's name "
                    "be displayed in the PDF, on the top of the first page "
                    "(lacking space).")
            raise PyBillProcessingException(
                                     text.encode("ascii", "xmlcharrefreplace"))



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main

from pybill.lib import ACCURACY
from pybill.lib.entities.accounting_docs import Downpayment

from generic_test_accounting_docs import AbstractAccDocTest


class DownpaymentTest(TestCase, AbstractAccDocTest):
    """
    Tests on the downpayment objects
    """
    def setUp(self):
        """
        Called before each test from this class
        """
        self.acc_doc = Downpayment("ID1")

    def test_empty_object(self):
        """
        Tests an empty downpayment: moreover the basic tests, tests the
        downpayment percent has the default value (30.0)
        """
        AbstractAccDocTest.test_empty_object(self)
        self.assertEqual(self.acc_doc.percent, 30.0)

    def test_get_total_without_holdback_to_be_paid(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback_free(),
                          round(242.112, ACCURACY))

    def test_get_total_holdback_to_be_paid(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback(),
                          round(0.0, ACCURACY))

    def test_get_total_vat_to_be_paid(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_vat(),
                          round(28.812, ACCURACY))

    def test_get_total_to_be_paid(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent.
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.get_to_be_paid(),
                          round(242.112, ACCURACY))

    def test_get_downpayment(self):
        """
        Tests the computation of the downpayment
        including taxes total * downpayment percent
        """
        self.build_standard_example()
        self.assertEqual(self.acc_doc.get_downpayment(),
                          round(242.112, ACCURACY))

    def test_get_total_without_holdback_to_be_paid_when_rounding(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent.
        """
        self.build_rounding_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback_free(),
                          round(50.23, ACCURACY))

    def test_get_total_holdback_to_be_paid_when_rounding(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent.
        """
        self.build_rounding_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback(),
                          round(0.0, ACCURACY))

    def test_get_total_vat_to_be_paid_when_rounding(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent.
        """
        self.build_rounding_example()
        self.assertEqual(self.acc_doc.get_to_be_paid_vat(),
                          round(8.232, ACCURACY))

    def test_get_total_to_be_paid_when_rounding(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent.
        """
        self.build_rounding_example()
        self.assertEqual(self.acc_doc.get_to_be_paid(),
                          round(50.23, ACCURACY))

    def test_get_downpayment_when_rounding(self):
        """
        Tests the computation of the downpayment
        including taxes total * downpayment percent
        """
        self.build_rounding_example()
        self.assertEqual(self.acc_doc.get_downpayment(),
                          round(50.23, ACCURACY))

    def test_get_total_without_holdback_to_be_paid_with_holdback(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent. The
        total without holdback to be paid is actually equal to this total
        except if it is greater than the including taxes total minus the total
        of holdbacks. In this case, the total without holdback to
        be paid is limited to (including taxes total - total
        of holdbacks) and the remaining sum is in total of holdbacks to be
        paid.
        """
        self.build_standard_example(with_holdbacks=True)
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback_free(),
                          round(242.112, ACCURACY))

    def test_get_total_holdback_to_be_paid_with_holdback(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent. The
        total without holdback to be paid is actually equal to this total
        except if it is greater than the including taxes total minus the total
        of holdbacks. In this case, the total without holdback to
        be paid is limited to (including taxes total - total
        of holdbacks) and the remaining sum is in total of holdbacks to be
        paid.
        """
        self.build_standard_example(with_holdbacks=True)
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback(),
                          round(0.0, ACCURACY))

    def test_get_total_vat_to_be_paid_with_holdback(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent.
        """
        self.build_standard_example(with_holdbacks=True)
        self.assertEqual(self.acc_doc.get_to_be_paid_vat(),
                          round(28.812, ACCURACY))

    def test_get_total_to_be_paid_with_holdback(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent.
        """
        self.build_standard_example(with_holdbacks=True)
        self.assertEqual(self.acc_doc.get_to_be_paid(),
                          round(242.112, ACCURACY))

    def test_get_downpayment_with_holdback(self):
        """
        Tests the computation of the downpayment when holdbacks are defined
        on some items
        including taxes total * downpayment percent
        """
        self.build_standard_example(with_holdbacks=True)
        self.assertEqual(self.acc_doc.get_downpayment(),
                          round(242.112, ACCURACY))

    def test_get_total_without_holdback_to_be_paid_with_huge_holdback(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent. The
        total without holdback to be paid is actually equal to this total
        except if it is greater than the including taxes total minus the total
        of holdbacks. In this case, the total without holdback to
        be paid is limited to (including taxes total - total
        of holdbacks) and the remaining sum is in total of holdbacks to be
        paid.

        Here, we have a huge holdback and thus we are in the second case.
        """
        self.build_standard_example(with_holdbacks=True)
        for i in [0,2,3]:
            self.acc_doc.acc_items[i].holdback_rate = 95.0
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback_free(),
                          round(138.772, ACCURACY))

    def test_get_total_holdback_to_be_paid_with_huge_holdback(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent. The
        total without holdback to be paid is actually equal to this total
        except if it is greater than the including taxes total minus the total
        of holdbacks. In this case, the total without holdback to
        be paid is limited to (including taxes total - total
        of holdbacks) and the remaining sum is in total of holdbacks to be
        paid.

        Here, we have a huge holdback and thus we are in the second case.
        """
        self.build_standard_example(with_holdbacks=True)
        for i in [0,2,3]:
            self.acc_doc.acc_items[i].holdback_rate = 95.0
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback(),
                          round(103.340, ACCURACY))

    def test_get_total_vat_to_be_paid_with_huge_holdback(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent.
        """
        self.build_standard_example(with_holdbacks=True)
        for i in [0,2,3]:
            self.acc_doc.acc_items[i].holdback_rate = 95.0
        self.assertEqual(self.acc_doc.get_to_be_paid_vat(),
                          round(28.812, ACCURACY))

    def test_get_total_to_be_paid_with_huge_holdback(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent.
        """
        self.build_standard_example(with_holdbacks=True)
        for i in [0,2,3]:
            self.acc_doc.acc_items[i].holdback_rate = 95.0
        self.assertEqual(self.acc_doc.get_to_be_paid(),
                          round(242.112, ACCURACY))

    def test_get_downpayment_with_huge_holdback(self):
        """
        Tests the computation of the downpayment when holdbacks are defined
        on some items
        including taxes total * downpayment percent
        """
        self.build_standard_example(with_holdbacks=True)
        for i in [0,2,3]:
            self.acc_doc.acc_items[i].holdback_rate = 95.0
        self.assertEqual(self.acc_doc.get_downpayment(),
                          round(242.112, ACCURACY))

    def test_get_total_without_holdback_to_be_paid_with_different_percent(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent. In this
        test, the downpayment percent is different from its default value
        """
        self.build_standard_example()
        self.acc_doc.percent = 50.0
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback_free(),
                          round(403.52, ACCURACY))

    def test_get_total_holdback_to_be_paid_with_different_percent(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent. In this
        test, the downpayment percent is different from its default value
        """
        self.build_standard_example()
        self.acc_doc.percent = 50.0
        self.assertEqual(self.acc_doc.get_to_be_paid_holdback(),
                          round(0.0, ACCURACY))

    def test_get_total_vat_to_be_paid_with_different_percent(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent. In this
        test, the downpayment percent is different from its default value
        """
        self.build_standard_example()
        self.acc_doc.percent = 50.0
        self.assertEqual(self.acc_doc.get_to_be_paid_vat(),
                          round(48.020, ACCURACY))

    def test_get_total_to_be_paid_with_different_percent(self):
        """
        In the specific case of a downpayment, the total to be paid is the
        including taxes total mutiplied with the downpayment percent. In this
        test, the downpayment percent is different from its default value
        """
        self.build_standard_example()
        self.acc_doc.percent = 50.0
        self.assertEqual(self.acc_doc.get_to_be_paid(),
                          round(403.52, ACCURACY))

    def test_get_downpayment_with_different_percent(self):
        """
        Tests the computation of the downpayment with a downpayment percent
        different from its default value.
        including taxes total * downpayment percent
        """
        self.build_standard_example()
        self.acc_doc.percent = 50.0
        self.assertEqual(self.acc_doc.get_downpayment(),
                          round(403.52, ACCURACY))


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-

from logilab.common.testlib import TestCase, unittest_main

from tempfile import mkdtemp
from shutil import copy
from os import remove, listdir, rmdir
from os import path as osp
from lxml import etree

from pybill.lib import PBD_XMLNS
from pybill.lib.entities.accounting_docs import Bill
from pybill.lib.xmlwriters.utils import ENCODING

from pybill.lib.controller import PyBillController


class FakeReader:
    """
    Class used to replace the real ``AccDocXMLReader``. We just need to
    checks the calls to the reader are correct.
    """
    def load_data(self, accdoc_file, cfg_file=None):
        """
        Fakes the data reading.
        """
        return "Read from %s with %s config" % (accdoc_file, cfg_file)


class FakeWriter:
    """
    Class used to replace the real ``PDFWriter``. We just need to
    check the calls to the writer are correct.
    """
    def __init__(self):
        self.writes = []

    def write(self, accdoc_entity, pdf_file):
        """
        Fakes the PDF building and writing.
        """
        self.writes.append("PDF write from %s in %s"
                           % (accdoc_entity, pdf_file))


class FakeEntriesGenerator:
    """
    Class used to replace the real ``EntriesGenerator``. We just need to
    check the calls to the generator are correct.
    """
    def __init__(self):
        self.actions = []

    def reset_entries(self):
        """
        Fakes the reset.
        """
        self.actions.append("Deletes all entries")
        
    def generate_entry(self, accdoc_entity):
        """
        Fakes the entry generation.
        """
        self.actions.append("Generates accounting entry for %s"
                            % accdoc_entity)

    def write_entries_in_xml(self, filename, encoding=ENCODING):
        """
        Fakes the XML writing.
        """
        self.actions.append("Writes entries in %s XML file in %s encoding" 
                            % (filename, encoding))



class ControllerTest(TestCase):
    def setUp(self):
        """
        Called before each test from this class.
        """
        self.controller = PyBillController()
        self.controller.accdoc_reader = FakeReader()
        self.controller.pdf_writer = FakeWriter()
        self.controller.entries_generator = FakeEntriesGenerator()
        self.temp_dir = mkdtemp()

    def tearDown(self):
        for fname in listdir(self.temp_dir):
            remove(osp.join(self.temp_dir, fname))
        rmdir(self.temp_dir)
        
    def test_build_entities_from_files_without_config(self):
        """
        Tests the ``build_entities_from_files`` method when no config
        file is specified.
        """
        filelist = [self.datapath("file1.xml"),
                    self.datapath("file2.xml"),
                    self.datapath("file3.xml")]
        result = self.controller.build_entities_from_files(filelist)
        self.assertEqual(result,
            [("Read from %s with None config" % self.datapath("file1.xml"),
              self.datapath("file1")),
             ("Read from %s with None config" % self.datapath("file2.xml"),
              self.datapath("file2")),
             ("Read from %s with None config" % self.datapath("file3.xml"),
              self.datapath("file3")),
            ])

    def test_build_entities_from_files_with_config(self):
        """
        Tests the ``build_entities_from_files`` method when a config
        file is specified.
        """
        filelist = [self.datapath("file1.xml"),
                    self.datapath("file2.xml"),
                    self.datapath("file3.xml")]
        cfgfile = self.datapath("config.xml")
        result = self.controller.build_entities_from_files(filelist,
                                                           cfgfile)
        self.assertEqual(result,
            [("Read from %s with %s config" % (self.datapath("file1.xml"),
                                               self.datapath("config.xml")),
              self.datapath("file1")),
             ("Read from %s with %s config" % (self.datapath("file2.xml"),
                                               self.datapath("config.xml")),
              self.datapath("file2")),
             ("Read from %s with %s config" % (self.datapath("file3.xml"),
                                               self.datapath("config.xml")),
              self.datapath("file3")),
            ])

    def test_write_pdf_to_files(self):
        """
        Tests the ``write_pdf_to_files`` method.
        """ 
        entitylist = [("entity1", "file1.pdf"),
                      ("entity2", "file2.pdf"),
                      ("entity3", "file3.pdf")]
        self.controller.write_pdf_to_files(entitylist)
        result = self.controller.pdf_writer.writes
        self.assertEqual(result,
                          ["PDF write from entity1 in file1.pdf",
                           "PDF write from entity2 in file2.pdf",
                           "PDF write from entity3 in file3.pdf"])

    def test_save_entities_to_xml_files(self):
        """
        Tests the ``save_entities_to_files`` method.
        """
        # Prepares the test
        for fname in listdir(self.datapath("accounting_dir")):
            copy(self.datapath("accounting_dir/%s" % fname),
                 osp.join(self.temp_dir, fname))
        # Does the actual test
        entitylist = [(Bill("id0"), osp.join(self.temp_dir, "accdoc0.xml")),
                      (Bill("id1"), osp.join(self.temp_dir, "accdoc1.xml")),
                      (Bill("id2"), osp.join(self.temp_dir, "accdoc2.xml")),
                      (Bill("id3"), osp.join(self.temp_dir, "accdoc3.xml"))]
        self.controller.save_entities_to_files(entitylist, ENCODING)
        files = listdir(self.temp_dir)
        files.sort()
        self.assertEqual(files, ["accdoc0.xml", "accdoc1.xml",
                                  "accdoc1.xml.old", "accdoc2.xml",
                                  "accdoc2.xml.old", "accdoc3.xml"])
        for filename, exp_id in [("accdoc0.xml", u"id0"),
                                 ("accdoc1.xml", u"id1"),
                                 ("accdoc1.xml.old", u"id1-prev"),
                                 ("accdoc2.xml", u"id2"),
                                 ("accdoc2.xml.old", u"id2-prev"),
                                 ("accdoc3.xml", u"id3")]:
            root = etree.parse(osp.join(self.temp_dir, filename)).getroot()
            doc_id = root.findtext(u"{%s}metadata/{%s}id"
                                                 % (PBD_XMLNS, PBD_XMLNS), u"")
            self.assertEqual(doc_id, exp_id)
        
    def test_export_accounting_entries_to_xml_file(self):
        """
        Tests the ``export_accounting_entries_to_xml_file`` method.
        """ 
        entitylist = ["entity1", "entity2", "entity3"]
        self.controller.export_accounting_entries_to_xml_file(entitylist,
                                                         "fileA.xml", ENCODING)
        result = self.controller.entries_generator.actions
        self.assertEqual(result,
                    ["Deletes all entries",
                     "Generates accounting entry for entity1",
                     "Generates accounting entry for entity2",
                     "Generates accounting entry for entity3",
                     "Writes entries in fileA.xml XML file in UTF-8 encoding"])


# definitions for automatic unit testing

if __name__ == '__main__':
    unittest_main()




# -*- coding: utf-8 -*-
"""
``pybill.lib.accounting.utils`` module contains some useful constants and the
basic classes used in the accounting entries generation.

The constants are used to designate the various account numbers used in the
accounting entries generation. These account numbers are stored in a dictionary
defined in each PyBill accounting entity.

.. autodata:: CLIENT

.. autodata:: CLIENT_HOLDBACK

.. autodata:: PRODUCT

.. autodata:: VAT

The basic class describes an accounting entry.

.. autoclass:: Entry
   :members:
   :undoc-members:
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"

from pybill.lib import ACCURACY
from pybill.lib.errors import PyBillProcessingException


CLIENT = u"Client"
"""
Constant used to designate, in the dictionaries, the account number of the
Client account used for an accounting document.

type: :class:`unicode`
"""

CLIENT_HOLDBACK = u"Client Holdback"
"""
Constant used to designate, in the dictionaries, the account number of the
Client Holdback account used for an accounting document.

type: :class:`unicode`
"""

PRODUCT = u"Product"
"""
Constant used to designate, in the dictionaries, the account number of the
Product account used for an accounting document.

type: :class:`unicode`
"""

VAT = u"VAT"
"""
Constant used to designate, in the dictionaries, the account number of the VAT
account used for an accounting document.

type: :class:`unicode`
"""


class Entry:
    """
    Little class representing an accounting entry.

    This entry is quite basic and contains only a date, an explanation, a list
    of debits and a list of credits.

    .. attribute:: date

       Attribute containing the date of the entry.

       type: :class:`datetime.date`

    .. attribute:: explanation

       Attribute containing the explanation of the entry.

       type: :class:`unicode`

    .. attribute:: credits

       Attribute containing a dictionary of account numbers (dictionary keys)
       associated with the amounts (floats) that must be credited on these
       accounts.

       type: dictionary of :class:`float` indexed by :class:`unicode` keys

    .. attribute:: debits

       Attribute containing a dictionary of account numbers (dictionary keys)
       associated with the amounts (floats) that must be debited on these
       accounts.

       type: dictionary of :class:`float` indexed by :class:`unicode` keys

    .. automethod:: __init__
    """
    def __init__(self, date, explanation):
        """
        Initializes a new accounting entry.

        :parameter date: Date of the entry.
        :type date: :class:`datetime.date`

        :parameter explanation: Explanation of the entry
        :type explanation: :class:`unicode`
        """
        self.date = date
        self.explanation = explanation
        self.credits = {}
        self.debits = {}

    def add_movement(self, account_number, amount):
        """
        Adds a movement of ``amount`` on the account whose number is
        ``account_number``.

        If the amount is positive, the account will be credited, else the
        account will be debited.

        :parameter account_number: Number of the accounting account that will
                                   be credited or debited in this entry.
        :type account_number: :class:`unicode`

        :parameter amount: Amount of the movement on the account
                           ``account_number``. If the amount is positive, the
                           account is credited, else it is debited.
        :type amount: :class:`float`
        """
        if amount >= 0:
            amts_list = self.credits.setdefault(account_number, [])
            amts_list.append(round(amount, ACCURACY))
        else:
            amts_list = self.debits.setdefault(account_number, [])
            amts_list.append(round(-amount, ACCURACY))

    def check(self):
        """
        Checks the accounting entry.

        That's to say, checks if the sum of the debits is equal to the sum
        of the credits. Raises an exception if it's not the case.
        """
        debit_amounts = 0.0
        for account, amounts in self.debits.items():
            debit_amounts += sum(amounts)
        credit_amounts = 0.0
        for account, amounts in self.credits.items():
            credit_amounts += sum(amounts)
        if abs(debit_amounts -credit_amounts) > 10**-(ACCURACY+3):
            raise PyBillProcessingException("Generated accounting entry %s "
                                            "has a different sum for its "
                                            "debits and its credits." % self)
        else:
            return True

    def __repr__(self):
        """
        :returns: A string representation of the accounting entry.
        :rtype: :class:`str`
        """
        str_repr = "<Entry "
        for account, amounts in self.credits.items():
            for amount in amounts:
                str_repr += "(Cred %s: %s) " % (account, amount)
        for account, amounts in self.debits.items():
            for amount in amounts:
                str_repr += "(Deb %s: %s) " % (account, amount)
        str_repr += ">"
        return str_repr



# -*- coding: utf-8 -*-
"""
``pybill.lib.pdfwriters``  package contains the classes and functions used to
produce PDF documents from the PyBill accounting documents.

This package uses :mod:`reportlab` library to build the PDF documents,
specifically the :mod:`~reportlab.platypus` library (see ReportLab documentation
for details).

From this package, the only public class that should be used outside of this
package is the :class:`~pybill.lib.pdfwriters.writers.PDFWriter` class. This
class can transform any PyBill accounting document into a PDF document. It is
imported and available in the namespace of this package.

Moreover, in this package, we define several `reportlab` paragraph styles that
will be used during the PDF generation.

.. data:: style_normal_left

   Object containing the style for a paragraph aligned on the left.

   type: :class:`reportlab.lib.styles.ParagraphStyle`

.. data:: style_normal_just

   Object containing the style for a paragraph justified on the left and the
   right.

   type: :class:`reportlab.lib.styles.ParagraphStyle`

.. data:: style_normal_right

   Object containing the style for a paragraph aligned on the right.

   type: :class:`reportlab.lib.styles.ParagraphStyle`

.. data:: style_normal_center

   Object containing the style for a centered paragraph.

   type: :class:`reportlab.lib.styles.ParagraphStyle`

.. data:: style_little_left

   Object containing the style for a paragraph aligned on the left but with a
   little font size.

   type: :class:`reportlab.lib.styles.ParagraphStyle`

.. data:: style_little_right

   Object containing the style for a paragraph aligned on the right but with a
   little font size.

   type: :class:`reportlab.lib.styles.ParagraphStyle`

.. data:: style_little_center

   Object containing the style for a centered paragraph but with a little font
   size.

   type: :class:`reportlab.lib.styles.ParagraphStyle`

.. data:: style_little_indent_left

   Object containing the style for a paragraph aligned on the left with a little
   font size and with a positive identation on the first line.

   type: :class:`reportlab.lib.styles.ParagraphStyle`

.. data:: style_footer

   Object containing the style for a paragraph of the footer (centered with 
   quite a little font size).

   type: :class:`reportlab.lib.styles.ParagraphStyle`

.. data:: style_title

   Object containing the style for a paragraph of the title (left aligned with a
   bigger font size).

   type: :class:`reportlab.lib.styles.ParagraphStyle`
"""
__docformat__ = "restructuredtext en"

from reportlab.lib.units import cm
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.enums import  TA_LEFT, TA_CENTER, TA_RIGHT, TA_JUSTIFY 


# Defining paragraph styles used in PDF generation

style_normal_left = ParagraphStyle("normal_left")
style_normal_left.fontSize = 11
style_normal_left.firstLineIndent = -1*cm
style_normal_left.leftIndent = 1*cm

style_normal_just = ParagraphStyle("normal_just")
style_normal_just.fontSize = 11
style_normal_just.firstLineIndent = -1*cm
style_normal_just.leftIndent = 1*cm
style_normal_just.alignment = TA_JUSTIFY

style_normal_right = ParagraphStyle("normal_right")
style_normal_right.fontSize = 11
style_normal_right.alignment = TA_RIGHT

style_normal_center = ParagraphStyle("normal_center")
style_normal_center.fontSize = 11
style_normal_center.alignment = TA_CENTER

style_little_left = ParagraphStyle("little_left")
style_little_left.fontSize = 9
style_little_left.firstLineIndent = -1*cm
style_little_left.leftIndent = 1*cm

style_little_right = ParagraphStyle("little_right")
style_little_right.fontSize = 9
style_little_right.alignment = TA_RIGHT

style_little_center = ParagraphStyle("little_center")
style_little_center.fontSize = 9
style_little_center.alignment = TA_CENTER

style_little_indent_left = ParagraphStyle("little_indent_left")
style_little_indent_left.fontSize = 9
style_little_indent_left.firstLineIndent = 0.5*cm

style_footer = ParagraphStyle("footer")
style_footer.fontSize = 10
style_footer.alignment = TA_CENTER

style_title = ParagraphStyle("title")
style_title.fontSize = 14
style_title.fontName = "Helvetica"
style_title.firstLineIndent = -1*cm
style_title.leftIndent = 1*cm


# Imports the class that must be used from the outside to write PDF bills
from pybill.lib.pdfwriters.writers import PDFWriter



# -*- coding: utf-8 -*-
"""
``pybill.lib.entities.accounting_docs.utils`` module defines some elementary
entities used by the entity classes.

.. autoclass:: AccItem
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: PreviousAccountingDoc
   :members:
   :undoc-members:
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"

from pybill.lib import ACCURACY


class AccItem:
    """
    Entity class representing an item that appears on the accounting document.

    .. attribute:: quantity

       Attribute defining the quantity of the item. Is not necessarily an
       integer, can be a real.

       type: :class:`float`

    .. attribute:: quantity_digits

       Attribute defining the number of digits to be written, for the quantity,
       on the documents where the item is displayed. This number of digits is
       also used to round the quantity before doing any of the item computations
       (VAT, price, etc.) By default, the number of digits is ``0``.

       type: :class:`int`

    .. attribute:: title:

       Attribute containing the description of the item

       type: :class:`unicode`

    .. attribute:: details

       Attribute defining a list of additional descriptions of the item.

       type: list of :class:`unicode`

    .. attribute:: unit_price

       Attribute defining the unit price of the item. Actually, this is a tax
       free price.

       type: :class:`float`

    .. attribute:: vat_rate

       Attribute defining the VAT rate that is applied to this item. This rate
       is ``None`` when no VAT is applicable to the item. Default value is
       ``None``. The rate is expressed in percents (e.g. ``19.6`` for 19.6%).

       type: :class:`float`

    .. attribute:: holdback_rate

       Attribute defining the holdback rate applied to this item. Hence,
       sometimes, an holdback is retained from the price of the item (and will
       be billed a few months later after the item has been fully tested). Can
       be ``None`` if no holdback is applied to this item. Default value is
       ``None``. The rate is expressed in percents (e.g. ``5.0`` for 5.0%).

       type: :class:`float`

    .. attribute:: holdback_on_vat

       Attribute defining a flag set to ``True`` if the holdback is also applied
       on VAT (thus the holback will be computed and substracted from the tax
       free price and from the VAT amount). When set to ``False``, the holdback
       is only applied on the tax free price (therefore, we substract the
       holdback from the tax free price but we keep the integrality of the VAT
       amount).

       For example: if the tax free price is ``100.0``, the VAT rate is ``20.0``
       and the holdback rate is ``5.0``. If the holdback is on VAT, we will pay:
       ``100.0*(1.0-0.05) + 20.0*(1.0-0.05) = 114.0``.  If the holdback is not
       on VAT, we will pay: ``100.0*(1.0-0.05) +20.0 = 115.0``.

       By default, the holdback is not applied on VAT (flag set to ``False``).

       type: :class:`bool`

    .. automethod:: __init__   
    """
    def __init__(self):
        """
        Initializes an empty object.
        """    
        self.quantity = 0.0
        self.quantity_digits = 0
        self.title = u""
        self.details = []
        self.unit_price = 0.0
        self.unit_price_digits = ACCURACY
        self.vat_rate = None
        self.holdback_rate = None
        self.holdback_on_vat = False

    def get_price(self):
        """
        Computes the price of this item (\ ``quantity`` * ``unit_price``\ ).

        This price is actually a tax free price.
        
        :returns: The price of the item.
        :rtype: :class:`float`
        """
        return round(round(self.quantity, self.quantity_digits) *
                     round(self.unit_price, self.unit_price_digits), ACCURACY)

    def get_vat_amount(self):
        """
        Computes the Value Added Tax (VAT) amount for this item
        (\ ``price`` * ``vat_rate``\ ).

        When no VAT is defined for the item, returns ``0.0``.
        
        :returns: The VAT amount of the item. ``0.0`` if no VAT rate is defined
                  for the item.
        :rtype: :class:`float`
        """
        if self.vat_rate is None:
            return 0.0
        else:
            return round(self.get_price() *
                         (round(self.vat_rate, ACCURACY) / 100.), ACCURACY)

    def get_including_taxes_price(self):
        """
        Computes the including taxes price for this item (\ ``price`` +
        ``vat_amount``\ ).

        When no VAT is defined for the item, returns the tax free price (\
        ``price``\ ).
        
        :returns: The including taxes price of the item or the price of the item
                  if no VAT rate is defined for the item.
        :rtype: :class:`float`
        """
        if self.vat_rate is None:
            return self.get_price()
        else:
            return round(self.get_price() + self.get_vat_amount(), ACCURACY)

    def get_holdback_amount(self):
        """    
        Computes the holdback amount for this item (\ ``price`` *
        ``holdback_rate``\ ).

        When no holdback rate is defined for the item, returns ``0.0``.
        
        :returns: The holdback amount of the item. ``0.0`` if no holdback rate
                  is defined for the item.
        :rtype: :class:`float`
        """
        if self.holdback_rate is None:
            return 0.0
        else:
            return round(self.get_price() *
                         (round(self.holdback_rate, ACCURACY) / 100.), ACCURACY)

    def get_holdback_vat_amount(self):
        """    
        Computes the holdback amount on the VAT amount for this item
        (\ ``vat_amount`` * ``holdback_rate``\ ).

        When no holdback rate is defined for the item, returns ``0.0``. When no
        VAT rate is defined for the item, returns ``0.0``.
        
        :returns: The holdback amount on the VAT amount of the item. ``0.0``
                  if no holdback rate or no VAT rate is defined for the item.
        :rtype: :class:`float`
        """
        if self.vat_rate is None:
            return 0.0
        elif self.holdback_rate is None:
            return 0.0
        elif self.holdback_on_vat is False:
            return 0.0
        else:
            return round(self.get_vat_amount() *
                         (round(self.holdback_rate, ACCURACY) / 100.), ACCURACY)


class PreviousAccountingDoc:
    """
    Entity class representing an accounting document that was previously sent
    and that is used in the current bill.

    This accounting document can be a downpayment that has previously been
    charged (``Downpayment`` previously sent) or a debit that has previously
    been issued (``Debit`` previously sent). The amount of this previous
    accounting document will modify the total to be paid of the current bill.

    The real type represented by a previous accounting document depends on the
    attribute where the object will be added in the current bill
    (``charged_downpayment`` or ``issued_debit``).

    .. attribute:: total

       Attribute containing the total of the previous accounting document.

       type: :class:`float`

    .. attribute:: vat_amount

       Attribute containing the VAT amount of the previous accounting
       document. This VAT amount is a part of the ``total`` previously defined.

       type: :class:`float`

    .. attribute:: accdoc_id

       Attribute defining the identifier of the previous accounting
       document. This identifier is not the document reference but the legal
       identifier of the document (cf.  ``id``)

       type: :class:`unicode`

    .. attribute:: accdoc_date

       Attribute defining the date when the previous accounting document was
       published.

       type: :class:`unicode`

    .. automethod:: __init__
    """
    def __init__(self):
        """
        Initializes an empty object.
        """        
        self.total = 0.0
        self.vat_amount = 0.0
        self.accdoc_id = u""
        self.accdoc_date = u""



# -*- coding: utf-8 -*-
"""
``pybill.lib.pdfwriters.flowgenerators.abstract`` module defines the abstract
base class used by all the PyBill-specific flow generators.

.. autoclass:: AbstractFlowGenerator
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"

from pybill.lib import ACCURACY
from reportlab.lib.units import cm
from reportlab.lib.colors import black

from reportlab.platypus import Paragraph, Spacer, KeepTogether
from reportlab.platypus import Table, TableStyle

from pybill.lib.pdfwriters import (style_title, style_normal_left,
                                   style_normal_center, style_normal_right,
                                   style_normal_just, style_little_indent_left,
                                   style_little_center, style_little_left)

from pybill.lib.pdfwriters.utils import TotalizableTable, NumberParagraph
from pybill.lib.pdfwriters.utils import escape_text, format_number


class AbstractFlowGenerator:
    """
    Abstract class containing the common parts for generating the PDF flow
    content for the various kinds of accounting documents (bill, claim form,
    downpayment, debit, pro-forma). Specific parts
    are generated in the  derived classes (one for each kind of document)
    (cf. `Template` design pattern).

    For a correct behaviour, you must use a flow generator corresponding to
    the kind of accounting document you want to render in PDF.
    
    The PDF flow content is a list of :class:`reportlab.platypus.Flowable`
    objects that will be displayed on various pages by the document
    template and the platypus machinery.

    .. automethod:: __init__

    .. automethod:: generate_flow

    .. automethod:: _get_acc_doc_title

    .. automethod:: _get_items_table
    
    .. automethod:: _get_bottom_left_element

    .. automethod:: _get_total_table

    .. automethod:: _get_total_table_cell_contents

    .. automethod:: _build_cell_contents_for_totals

    """
    def __init__(self):
        """
        Initializes the flow generator.

        This constructor should never be called directly as this class is
        abstract.
        """
        if self.__class__ is AbstractFlowGenerator:
            raise NotImplementedError("Abstract class can't be instanciated")

    def generate_flow(self, acc_doc):
        """
        Generates the flow content (a list of
        :class:`~reportlab.platypus.Flowable`) for the accounting
        document.
        
        :parameter acc_doc: The current accounting document that must be
                            exported in PDF.
        :type acc_doc:
            :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`

        :returns: A list of flowable objects to be displayed by the PDF
                  document template.
        :rtype: list of :class:`reportlab.platypus.Flowable`
        """
        flow_list = []
        # Title.
        flow_list.append(Paragraph(self._get_acc_doc_title(acc_doc),
                                   style_title))
        # Introduction text.
        txt = acc_doc.cfg_data.get_local_phrase(u'intro-detail')
        if txt != u"":
            flow_list.append(Spacer(1, 0.1*cm))
            flow_list.append(Paragraph(u"<i>%s</i>" % escape_text(txt),
                                       style_little_left))
        # Optional remarks.
        for rem in [r.strip() for r in acc_doc.remarks if r.strip() != u""]:
            flow_list.append(Spacer(1, 0.2*cm))
            flow_list.append(Paragraph(escape_text(rem, True),
                                       style_normal_just))
        # End of the top part of the document.
        flow_list.append(Spacer(1, 0.5*cm))
        # Items table (main table).
        flow_list.append(self._get_items_table(acc_doc))
        # End of the body part of the document.
        flow_list.append(Spacer(1, 0.5*cm))
        # Bottom table containing en element on the left (table of payment
        # terms or set of paragraphs for the pro-forma) and the total table
        # on the right.
        bott_table = Table([[self._get_bottom_left_element(acc_doc),
                             u" ",
                             self._get_total_table(acc_doc)]],
                           [9*cm, 2.5*cm, 7.5*cm], None)
        bott_table.setStyle(
            TableStyle([ ('ALIGN',         (0,0), ( 0,-1), 'LEFT'),
                         ('ALIGN',         (2,0), ( 2,-1), 'RIGHT'),
                         ('VALIGN',        (0,0), (-1,-1), 'TOP'),
                         ('LEFTPADDING',   (0,0), (-1,-1), 0),
                         ('RIGHTPADDING',  (0,0), (-1,-1), 0),
                         ('TOPPADDING',    (0,0), (-1,-1), 0),
                         ('BOTTOMPADDING', (0,0), (-1,-1), 0),
                       ])
            )
        # The table is in a ``KeepTogether`` because we don't want it to be
        # splitted on two pages.
        flow_list.append(KeepTogether([bott_table]))
        return flow_list

    def _get_acc_doc_title(self, acc_doc):
        """
        Gives the title of the accounting document.

        The title depends on the kind of document. This abstract method will be
        defined in the children classes.

        :parameter acc_doc: The current accounting document that must be
                            exported in PDF.
        :type acc_doc:
            :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`

        :returns: The accounting document title.
        :rtype: :class:`unicode`
        """
        raise NotImplementedError()

    def _get_items_table(self, acc_doc):
        """
        Gives the main table of the accounting document that details the
        items.

        The table can be a 5 columns table (quantity, description, vat
        rate, tax free unit price, tax free price) or a 4 columns table
        (quantity, description, tax free unit price, tax free price).

        The table is a :class:`~pybill.lib.pdfwriters.utils.TotalizableTable`
        instead of a classic :class:`reportlab.platypus.Table` because we need
        this specific table to compute the carry forwards in the
        :class:`~pybill.lib.pdfwriters.templates.doc_templates.AccDocDocumentTemplate`
        document template. Inside this specific table, in each row, we insert
        one :class:`~pybill.lib.pdfwriters.utils.NumberParagraph` that contains
        the total of the row. This allows us to know what to add in the carry
        forward.
        
        :parameter acc_doc: The current accounting document that must be
                            exported in PDF.
        :type acc_doc:
            :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`

        :returns:
            The table detailling the items. This table is a
            :class:`~pybill.lib.pdfwriters.utils.TotalizableTable` to allow the
            computation of the carry forwards in the document template. The
            items table have 4 columns when the accounting document has no VAT
            rate and 5 columns when at least one item of the accounting document
            has a VAT rate.
        :rtype: :class:`~pybill.lib.pdfwriters.utils.TotalizableTable`
        """
        cfg_data = acc_doc.cfg_data
        table_data = []
        # Header row.
        row = []
        row.append(Paragraph(
                           escape_text(cfg_data.get_local_phrase(u'quantity')),
                             style_little_center))
        row.append(Paragraph(
                        escape_text(cfg_data.get_local_phrase(u'description')),
                             style_little_center))
        if acc_doc.has_vat_rates():
            row.append(Paragraph(
                           escape_text(cfg_data.get_local_phrase(u'vat-rate')),
                                 style_little_center))
            row.append(Paragraph(
                      escape_text(cfg_data.get_local_phrase(u'tf-unit-price')),
                                 style_little_center))
            row.append(Paragraph(
                           escape_text(cfg_data.get_local_phrase(u'tf-price')),
                                 style_little_center))
        else:
            row.append(Paragraph(
                         escape_text(cfg_data.get_local_phrase(u'unit-price')),
                                 style_little_center))
            row.append(Paragraph(
                              escape_text(cfg_data.get_local_phrase(u'price')),
                                 style_little_center))            
        table_data.append(row)
        # Item rows.
        for item in acc_doc.acc_items:
            row = []
            # First cell: quantity.
            row.append(Paragraph(format_number(item.quantity,
                                               item.quantity_digits,
                                               cfg_data.number_separators),
                                 style_little_center) )
            # Second cell: description (list of paragraphs).
            cell = []
            cell.append(Paragraph(escape_text(item.title, True),
                                  style_normal_left))
            for detail in item.details:
                cell.append(Paragraph(u"<i>%s</i>" % escape_text(detail, True),
                                      style_little_indent_left))
            if item.holdback_rate is not None:
                cell.append(Spacer(0.1*cm, 1))
                text = u"%s " % cfg_data.get_local_phrase(u"holdback-on")
                if item.holdback_on_vat and item.vat_rate is not None:
                    text += cfg_data.get_local_phrase(u"it-price")
                elif item.vat_rate is not None:
                    text += cfg_data.get_local_phrase(u"tf-price")
                else:
                    text += cfg_data.get_local_phrase(u"price")
                text += u"%s %s%% %s %s" \
                        % (cfg_data.get_local_phrase(u"colon"),
                           format_number(item.holdback_rate, ACCURACY,
                                         cfg_data.number_separators),
                           cfg_data.get_local_phrase(u"ita-est"),
                           format_number(item.get_holdback_amount(), ACCURACY,
                                         cfg_data.number_separators))
                if item.holdback_on_vat and item.vat_rate is not None:
                    text += " + %s" \
                            % (format_number(item.get_holdback_vat_amount(),
                                         ACCURACY, cfg_data.number_separators))
                cell.append(Paragraph(escape_text(text), style_normal_left))
            row.append(cell)
            # Third cell: VAT rate (only if the document has VAT rates).
            if acc_doc.has_vat_rates():
                if item.vat_rate is not None:
                    text = u"%s%%" % format_number(item.vat_rate, ACCURACY,
                                                   cfg_data.number_separators)
                else:
                    text = u""
                row.append(Paragraph(text, style_normal_right))
            # Third/Fourth cell: tax free unit price.
            row.append(Paragraph(format_number(item.unit_price,
                           item.unit_price_digits, cfg_data.number_separators),
                                 style_normal_right))
            # Fourth/Fifth cell: tax free price (total of the row).
            # We don't forget to use a ``NumberParagraph`` in order to be
            # able to process the carry forward with the ``TotalizableTable``.
            row.append(NumberParagraph(item.get_price(),
                                       format_number(item.get_price(), ACCURACY,
                                                   cfg_data.number_separators),
                                       style_normal_right))
            table_data.append(row)
        if acc_doc.has_vat_rates():
            # 5 columns table.
            table_dims = [1.5*cm, 10*cm, 2*cm, 2.5*cm, 3*cm]
        else:
            # 4 columns table.
            table_dims = [1.5*cm, 12*cm, 2.5*cm, 3*cm]
        # We use a ``TotalizableTable`` in order to be able to compute the
        # carry forward with the ``AccDocDocumentTemplate`` document template.
        table = TotalizableTable(table_data, table_dims, None)
        table.repeatRows = 1 # When splitting the table, repeats first row.
        table.setStyle(
            TableStyle([ ('ALIGN',        (0,0), (-1, 0), 'CENTER'),
                         ('ALIGN',        (0,1), ( 0,-1), 'RIGHT'),
                         ('ALIGN',        (1,1), ( 1,-1), 'LEFT'),
                         ('ALIGN',        (2,1), (-1,-1), 'RIGHT'),
                         ('VALIGN',       (0,0), (-1,-1), 'TOP'),
                         ('BOX',          (0,0), (-1,-1), 0.5, black),
                         ('BOX',          (0,0), (-1, 0), 0.5, black),
                         ('INNERGRID',    (0,0), (-1,-1), 0.25,black),
                         ('LEFTPADDING',  (0,0), (-1, 0), 3),
                         ('RIGHTPADDING', (0,0), (-1, 0), 3),
                       ])
                      )
        return table

    def _get_bottom_left_element(self, acc_doc):
        """
        Gives the element that must be displayed at the bottom left of
        the accounting document.

        This element depends on the kind of document, but is actually a payment
        terms table. This method can be redefined in the children classes.
        
        :parameter acc_doc: The current accounting document that must be
                            exported in PDF.
        :type acc_doc:
            :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`

        :returns: A flowable  object contaning the content to be displayed
                  in the bottom left of the PDF document.
        :rtype: :class:`reportlab.platypus.Flowable`
        """
        table_data = []
        # First row contains a single cell with the payment terms.
        row = []
        row.append(Paragraph(
              escape_text(acc_doc.cfg_data.get_local_phrase(u'payment-terms')),
                             style_normal_center))
        table_data.append(row)
        #second row contains a single cell with several paragraphs (bank data).
        cell = [] 
        if acc_doc.payment_terms.strip() != u"":
            cell.append(Paragraph(escape_text(acc_doc.payment_terms, True),
                                  style_normal_left))
        for bk_line in acc_doc.cfg_data.bank_data_lines:
            cell.append(Paragraph(escape_text(bk_line), style_normal_left))
        table_data.append([cell])
        # Builds the payment terms table (with its style).
        pt_table = Table(table_data, [9*cm], None)
        pt_table.setStyle(
            TableStyle([ ('ALIGN',     (0,0), (-1, 0), 'CENTER'),
                         ('ALIGN',     (0,1), (-1, 1), 'LEFT'),
                         ('VALIGN',    (0,0), (-1,-1), 'TOP'),
                         ('BOX',       (0,0), (-1,-1), 0.50, black),
                         ('INNERGRID', (0,0), (-1,-1), 0.25, black),
                       ])
                         )
        return pt_table

    def _get_total_table(self, acc_doc):
        """
        Gives the total table of the accounting document (this table appears at
        the bottom right of the PDF document).

        The content of this table depends on the kind of document.
        
        This method relies on the :meth:`_get_total_table_cell_contents` that
        gives the contents of the cells of this table. This method builds
        a :class:`~reportlab.platypus.Table` object with these contents.
        
        :parameter acc_doc: The current accounting document that must be
                            exported in PDF.
        :type acc_doc:
            :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`

        :returns: The total table. This table embeds its style and is ready to
                  be inserted in the flow.
        :rtype: :class:`reportlab.platypus.Table`
        """
        table_data = []
        for name, number, style in \
                        self._get_total_table_cell_contents(acc_doc):
            start_tag = end_tag = u""
            if u"bold" in style.split():
                start_tag += u"<b>"
                end_tag = u"</b>" + end_tag
            if u"italic" in style.split():
                start_tag += u"<i>"
                end_tag = u"</i>" + end_tag
            row = []
            if name is None or name == []:
                cell = [Paragraph(u"", style_normal_left)]
            else:
                cell = [Paragraph(u"%s%s%s" %(start_tag,
                                              escape_text(name[0], True),
                                              end_tag),
                                  style_normal_left)]
                cell.extend([Paragraph(u"%s%s%s" %(start_tag,
                                                   escape_text(line, True),
                                                   end_tag),
                                       style_little_left)
                             for line in name[1:]])
            row.append(cell)
            row.append(Paragraph("%s%s%s" %(start_tag,
                                            format_number(number, ACCURACY,
                                           acc_doc.cfg_data.number_separators),
                                            end_tag),
                                 style_normal_right))
            table_data.append(row)
        table = Table(table_data, [4.5*cm, 3*cm], None)
        table.setStyle(
            TableStyle([ ('ALIGN',     (0,0), ( 0,-1), 'LEFT'),
                         ('ALIGN',     (1,0), ( 1,-1), 'RIGHT'),
                         ('VALIGN',    (0,0), (-1,-1), 'MIDDLE'),
                         ('BOX',       (0,0), (-1,-1), 0.50, black),
                         ('INNERGRID', (0,0), (-1,-1), 0.25, black),
                       ])
                      )
        return table

    def _get_total_table_cell_contents(self, acc_doc):
        """
        Gives the contents of the cells of the total table.

        These contents will be used by the :meth:`_get_total_table` method to
        build the total table.  The cell contents depend on the kind of
        accounting document currently processed.
        
        The contents of the cells are given as a list of triples: name,
        number and style. Each couple corresponds to one row of the total
        table.

        - The name is a list of unicode strings corresponding to several
          lines describing the content of the row. It is written in the first
          cell of the row. The first line is written with the regular font,
          the next are written with littler font.

        - The number is a float corresponding to the amount of the row. It is
          written in the second cell of the row.

        - The style is a specification of how the row must be displayed. It can
          contain u"bold", u"italic", both or none of them.

        Please note that the name will be escaped when inserted in the table.
        Therefore it should not contain any tag of style. You should use the
        style element in the triple to specify the style of the row.

        This method is abstract and will be implemented in the child classes.
        
        :parameter acc_doc: The current accounting document that must be
                            exported in PDF.
        :type acc_doc:
            :class:`~pybill.lib.entities.accounting_docs.abstract.GenericAccountingDoc`

        :returns:
            A list describing the cells of the total table. The first element is
            a list of unicode strings corresponding to the various lines of the
            name of the row; the second element is a float corresponding to the
            number written in the row; the third element is a specification of
            the style of the row (``u"bold"``, ``u"italic"``,
            ``u"bold italic"``, ``u""``). 
        :rtype: list of triples (list of :class:`unicode`, :class:`float`,
                :class:`unicode`)
        """
        raise NotImplementedError()

    def _build_cell_contents_for_totals(self, acc_doc):
        """
        Builds the cell contents corresponding to the first rows of the total
        table.

        These rows describe the totals (tax free total, vat amounts,
        including taxes total, holdbacks). As they are inserted in the total
        tables generated by all the subclasses, they have been factorized here.

        See :meth:`_get_total_table_cell_contents` method for a description of
        the triples describing the cell contents in the total table.

        This method generates different rows if the accounting document contain
        or doesn't contain VAT rates, contain or doesn't contain holdbacks. If
        it has got VAT rates, the totals rows will be: tax free total, various
        VAT amounts, including taxes total. If it hasn't got VAT rates, the
        totals rows will be: total. If it has holdbacks, two or three more rows
        will be added: holdbacks on tax free amounts and holdbacks on VAT
        amounts (if the holdbacks are on VAT) and holdbacks to be paid.

        :returns:        
            A list describing the first cells of the total table, corresponding
            to the totals. The first element is a list of unicode strings
            corresponding to the various lines of the name of the row; the
            second element is a float corresponding to the number written in the
            row; the third element is a specification of the style of the row
            (``u"bold"``, ``u"italic"``, ``u"bold italic"``, ``u""``). 
        :rtype: list of triples (list of :class:unicode`, :class:`float`,
                :class:`unicode`)
        """
        cfg_data = acc_doc.cfg_data
        contents = []
        if not acc_doc.has_vat_rates():
            # Total.
            name = [cfg_data.get_local_phrase(u'total')]
            number = acc_doc.get_price_total()
            contents.append((name, number, u"bold"))
            if acc_doc.has_holdbacks():
                name = [u"%s %s" % (cfg_data.get_local_phrase(u'including'),
                                    cfg_data.get_local_phrase(u'holdback')),]
                number = acc_doc.get_holdback_amount()
                contents.append((name, number, u"italic"))
        else:
            # Tax free total.
            name = [cfg_data.get_local_phrase(u'tf-total')]
            number = acc_doc.get_price_total()
            contents.append((name, number, u"regular"))
            # VAT amounts.
            for vat_rate in acc_doc.get_all_vat_rates():
                name = [u"%s %s%%"
                        % (cfg_data.get_local_phrase(u'vat-amount'),
                           format_number(vat_rate, ACCURACY,
                                                  cfg_data.number_separators))]
                number = acc_doc.get_vat_amount_for_rate(vat_rate)
                contents.append((name, number, u"regular"))
            # Including taxes total.
            name = [cfg_data.get_local_phrase(u'it-total')]
            number = acc_doc.get_including_taxes_total()
            contents.append((name, number, u"bold"))
            if acc_doc.has_holdbacks():
                name = [u"%s %s" % (cfg_data.get_local_phrase(u'including'),
                                    cfg_data.get_local_phrase(u'holdback')),
                        cfg_data.get_local_phrase(u'on-tf')]
                number = acc_doc.get_holdback_amount()
                contents.append((name, number, u"italic"))
                if acc_doc.has_vat_holdbacks():
                    name = [u"%s %s" % (cfg_data.get_local_phrase(u'including'),
                                        cfg_data.get_local_phrase(u'holdback')),
                            cfg_data.get_local_phrase(u'on-vat')]
                    number = acc_doc.get_holdback_vat_amount()
                    contents.append((name, number, u"italic"))
        return contents



# -*- coding: utf-8 -*-

from pybill.lib.entities.accounting_docs import AccItem
from pybill.lib.entities.addresses import PersonAddress, OrganisationAddress
import datetime

class FormatWriterGenericTest:
    """
    Abstract class containing various useful methods for testing accounting
    documents writing in XML.
    """
    def __init__(self) :
        if self.__class__ is AccDocXMLWriterGenericTest:
            raise NotImplementedError()

    def _fill_accdoc(self):
        """
        Method called after ``self.accdoc`` has been built in the ``setUp``
        method of the test cases. This method fills the accounting document
        with common content.
        """
        self.accdoc.doc_ref = u"A2"
        self.accdoc.place = u"A3"
        self.accdoc.date = u"A4"
        self.accdoc.date_num = datetime.date(2009, 10, 19)
        for i in range(5, 11, 2):
            self.accdoc.other_infos[u"A%d" % i] = u"A%d" % (i+1)
        self.accdoc.sender = self._build_address(u"B")
        self.accdoc.receiver = self._build_address(u"C")
        for i in range(1, 3):
            self.accdoc.remarks.append(u"D%d" % i)
        for key, num in [(u"E", 5), (u"F", 6), (u"G", 7), (u"H", 8)]:
            acc_item = AccItem()
            acc_item.quantity = (num * 100.0 + 1)
            acc_item.title = u"%s2" % key
            acc_item.unit_price = (num * 100.0 + 3)
            acc_item.vat_rate = (num * 100.0 + 4)
            self.accdoc.acc_items.append(acc_item)
        self.accdoc.acc_items[0].quantity_digits = 5
        self.accdoc.acc_items[0].details = [u"E3", u"E4"]
        self.accdoc.acc_items[0].unit_price = 505.0
        self.accdoc.acc_items[0].vat_rate = None
        self.accdoc.acc_items[1].unit_price_digits = 6
        self.accdoc.acc_items[2].holdback_rate = 7.0
        self.accdoc.acc_items[3].holdback_rate = 8.0
        self.accdoc.acc_items[3].holdback_on_vat = True

    def _build_address(self, prefix):
        """
        Builds a ``PersonAddress`` object suitable for the tests. The testing
        values in this object are prefixed with ``prefix``.
        """
        addr = PersonAddress()
        addr.honorific = u"%s1" % prefix
        addr.firstname = u"%s2" % prefix
        addr.other_name = u"%s3" % prefix
        addr.surname  = u"%s4" % prefix
        addr.lineage = u"%s5" % prefix
        self._fill_address(addr, prefix, 6)
        addr.organisation = org = OrganisationAddress()
        org.name = u"%s17" % prefix
        org.divisions = [u"%s18" % prefix, u"%s19" % prefix]
        org.job_titles = [u"%s20" % prefix, u"%s21" % prefix]
        self._fill_address(org, prefix, 22)
        return addr

    def _fill_address(self, addr, prefix, start_num):
        """
        Fills the ``address`` object with common pieces of information. The
        values of these pieces of information are prefixed with ``prefix`` and
        contain a number starting at ``start_num``.
        """
        addr.streets = [u"%s%d" % (prefix, start_num + 0),
                        u"%s%d" % (prefix, start_num + 1)]
        addr.post_box = u"%s%d" % (prefix, start_num + 2)
        addr.post_code = u"%s%d" % (prefix, start_num +3)
        addr.city = u"%s%d" % (prefix, start_num + 4)
        addr.state = u"%s%d" % (prefix, start_num + 5)
        addr.country = u"%s%d" % (prefix, start_num + 6)
        addr.phone = u"%s%d" % (prefix, start_num +7)
        addr.fax = u"%s%d" % (prefix, start_num + 8)
        addr.web = u"%s%d" % (prefix, start_num + 9)
        addr.email = u"%s%d" % (prefix, start_num + 10)
